# coding=utf-8
import datetime
import json
import os
import re

import mock
import pandas as pd
import pytest
import responses
import six

from datarobot import (
    AdvancedOptions,
    AUTOPILOT_MODE,
    Blueprint,
    errors,
    Featurelist,
    Model,
    ModelJob,
    PredictJob,
    Project,
    QUEUE_STATUS,
    RatingTable,
    RatingTableModel,
    SharingAccess,
    TARGET_TYPE,
    UserCV,
    UserTVH,
)
from datarobot.enums import PROJECT_STAGE, SCALEOUT_MODELING_MODE
from tests.test_helpers import fixture_file_path, URLParamsTestCase
from tests.utils import assert_raised_regex, assert_urls_equal, request_body_to_json, SDKTestcase

if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch


# FIXTURES and HELPERS #############################################################################


def construct_dummy_aimed_project(
    pid="p-id", project_name="Untitled Project", target="SalePrice", autopilot_mode=0,
):
    return {
        "id": pid,
        "projectName": project_name,
        "fileName": "Untitled Project.csv",
        "stage": "modeling",
        "autopilotMode": autopilot_mode,
        "created": "2015-05-21T16:02:02.573565",
        "target": target,
        "metric": "Gamma Deviance",
        "partition": {
            "cvMethod": "random",
            "validationType": "CV",
            "validationPct": None,
            "holdoutPct": 20,
            "reps": 5,
            "holdoutLevel": None,
            "validationLevel": None,
            "trainingLevel": None,
            "partitionKeyCols": None,
            "userPartitionCol": None,
            "cvHoldoutLevel": None,
            "datetimeCol": None,
        },
        "recommender": {
            "isRecommender": False,
            "recommenderItemId": None,
            "recommenderUserId": None,
        },
        "advancedOptions": {
            "weights": None,
            "blueprintThreshold": None,
            "responseCap": False,
            "seed": None,
            "smartDownsampled": False,
            "majorityDownsamplingRate": None,
            "defaultMonotonicIncreasingFeaturelistId": None,
            "defaultMonotonicDecreasingFeaturelistId": None,
            "offset": None,
            "exposure": None,
            "allowedPairwiseInteractionGroups": [["A", "B", "C"], ["C", "D"]],
        },
        "positiveClass": None,
        "maxTrainPct": 64.0,
        "maxTrainRows": 128,
        "scaleoutMaxTrainPct": 64.0,
        "scaleoutMaxTrainRows": 128,
        "holdoutUnlocked": True,
        "targetType": "Regression",
        "unsupervisedMode": True,
    }


AIMED_PROJECT_JSON = json.dumps(construct_dummy_aimed_project())


def prep_successful_project_cloning_responses(new_name=None):
    """
    Setup the responses library to mock calls to the API necessary to
    clone a project

    Returns
    ----------
    source_pid : str
        ID of the source project to be cloned
    """
    source_project = construct_dummy_aimed_project()
    project = construct_dummy_aimed_project(
        project_name=new_name or "Copy of {}".format(source_project["projectName"]), pid="id2",
    )
    pid = project.get("id")
    responses.add(
        responses.POST,
        "https://host_name.com/projectClones/",
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )
    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": "https://host_name.com/projects/{}/".format(pid)},
    )
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/".format(pid),
        status=200,
        body=json.dumps(project),
        content_type="application/json",
    )
    source_pid = source_project.get("id")
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/".format(source_pid),
        status=200,
        body=json.dumps(source_project),
        content_type="application/json",
    )
    return source_pid


def prep_successful_project_creation_responses(project=None):
    """
    Setup the responses library to mock calls to the API necessary to
    create a project

    Parameters
    ----------
    project : dict
        Project JSON as a dict, as would be expected from construct_dummy_aimed_project_json
    """
    if project is None:
        project = construct_dummy_aimed_project()

    pid = project.get("id")
    responses.add(
        responses.POST,
        "https://host_name.com/projects/",
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )
    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": "https://host_name.com/projects/{}/".format(pid)},
    )
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/".format(pid),
        status=200,
        body=json.dumps(project),
        content_type="application/json",
    )


def prep_successful_aim_responses(project=None):
    """A helper to use with setting up test scenarios where the server is
    expected to successfully set the target.

    Parameters
    ----------
    project : dict
        Project JSON as a dict, as would be expected from construct_dummy_aimed_project_json
    """
    if project is None:
        project = construct_dummy_aimed_project()

    pid = project.get("id")
    project_url = "https://host_name.com/projects/{}/".format(pid)
    aim_url = project_url + "aim/"

    responses.add(
        responses.PATCH,
        aim_url,
        body="",
        status=202,
        adding_headers={"Location": "https://host_name.com/status/some-status/"},
        content_type="application/json",
    )

    responses.add(
        responses.GET,
        "https://host_name.com/status/some-status/",
        body="",
        status=303,
        adding_headers={"Location": project_url},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        project_url,
        body=json.dumps(project),
        status=200,
        content_type="application/json",
    )


# TESTS ############################################################################################


def test_future_proof(project_with_target_data):
    Project.from_data(dict(project_with_target_data, future="new"))


@pytest.mark.usefixtures("known_warning")
def test_project_from_dict_is_deprecated(project_without_target_data):
    Project(project_without_target_data)


class ProjectTestCase(SDKTestcase):
    def setUp(self):
        self.blueprint_api_resp = [
            {
                "projectId": "p-id",
                "processes": [
                    "Regularized Linear Model Preprocessing v5",
                    "Log Transformed Response",
                ],
                "id": "cbb4d6101dea1768ed79d75edd84c6c7",
                "modelType": "Auto-tuned K-Nearest Neighbors Regressor (Minkowski Distance)",
            },
            {
                "projectId": "p-id",
                "processes": [
                    "Regularized Linear Model Preprocessing v12",
                    "Log Transformed Response",
                ],
                "id": "e0708bd47f9fd21019a5ab7846e8364d",
                "modelType": "Auto-tuned Stochastic Gradient Descent Regression",
            },
        ]

    def test_instantiation_with_data(self):
        """
        Test instantiation Project(data)
        """
        data = {
            "id": "project_id",
            "project_name": "project_test_name",
            "mode": 2,
            "stage": "stage",
            "target": "test_target",
        }
        project_inst = Project.from_data(data)
        self.assertEqual(project_inst.id, data["id"])
        self.assertEqual(project_inst.project_name, data["project_name"])
        self.assertEqual(project_inst.mode, data["mode"])
        self.assertEqual(project_inst.stage, data["stage"])
        self.assertEqual(project_inst.target, data["target"])

        self.assertEqual(repr(project_inst), "Project(project_test_name)")

    def test_print_project_nonascii_name(self):
        project = Project.from_data(
            {"id": "project-id", "project_name": u"\u3053\u3093\u306b\u3061\u306f"}
        )
        print(project)

    def test_get_permalink(self):
        p = Project("pid")
        expected = "https://host_name.com/projects/pid/models"
        self.assertEqual(expected, p.get_leaderboard_ui_permalink())

    def test_set_project_description(self):
        description = "This is my test project."
        p = Project("pid", project_description=description)
        assert p.project_description == description

    @responses.activate
    def test_update_project_description(self):
        description = "This is my test project."
        updated_description = "Project XYZ"
        responses.add(
            responses.PATCH,
            "https://host_name.com/projects/p-id/",
            body="",
            status=200,
            content_type="application/json",
        )
        p = Project("p-id", project_description=description)
        p.set_project_description(updated_description)
        assert responses.calls[0].request.method == "PATCH"
        payload = request_body_to_json(responses.calls[0].request)
        assert p.project_description == payload["projectDescription"] == updated_description

    @mock.patch("webbrowser.open")
    def test_open_leaderboard_browser(self, webbrowser_open):
        project = Project("p-id")
        project.open_leaderboard_browser()
        self.assertTrue(webbrowser_open.called)

    @responses.activate
    def test_create_project_async(self):
        prep_successful_project_creation_responses()

        new_project = Project.create(sourcedata="https://google.com")
        self.assertEqual(new_project.id, "p-id")
        self.assertEqual(new_project.project_name, "Untitled Project")

    @responses.activate
    def test_create_project_non_ascii_async(self):
        prep_successful_project_creation_responses()
        name = u"\xe3\x81\x82\xe3\x81\x82\xe3\x81\x82"
        Project.create("https://google.com", project_name=name)

    @responses.activate
    def test_get_project_metrics(self):
        """
        Test get project metrics
        """
        raw = """
        {"available_metrics":
            ["Gini Norm",
              "Weighted Gini Norm",
              "Weighted R Squared",
              "Weighted RMSLE",
              "Weighted MAPE",
              "Weighted Gamma Deviance",
              "Gamma Deviance",
              "RMSE",
              "Weighted MAD",
              "Tweedie Deviance",
              "MAD",
              "RMSLE",
              "Weighted Tweedie Deviance",
              "Weighted RMSE",
              "MAPE",
              "Weighted Poisson Deviance",
              "R Squared",
              "Poisson Deviance"],
         "feature_name": "SalePrice"}
        """
        expected_url = "https://host_name.com/projects/p-id/features/metrics/"
        responses.add(
            responses.GET, expected_url, body=raw, status=200, content_type="application/json"
        )
        get_project = Project("p-id").get_metrics("SalePrice")
        assert_urls_equal(responses.calls[0].request.url, expected_url + "?featureName=SalePrice")
        self.assertEqual(get_project["feature_name"], "SalePrice")

    @responses.activate
    def test_get_project(self):
        """
        Test get project
        """
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/",
            body=AIMED_PROJECT_JSON,
            status=200,
            content_type="application/json",
        )
        get_project = Project.get("p-id")
        assert get_project.id == "p-id"
        assert get_project.project_name == "Untitled Project"
        assert get_project.target == "SalePrice"
        assert get_project.target_type == "Regression"
        assert get_project.stage == "modeling"
        assert get_project.metric == "Gamma Deviance"
        assert get_project.positive_class is None
        assert get_project.max_train_pct == 64.0
        assert get_project.max_train_rows == 128
        assert get_project.scaleout_max_train_pct == 64.0
        assert get_project.scaleout_max_train_rows == 128
        assert get_project.holdout_unlocked is True
        assert get_project.unsupervised_mode is True
        assert isinstance(get_project.partition, dict)
        assert isinstance(get_project.recommender, dict)
        assert isinstance(get_project.advanced_options, dict)
        assert "weights" not in get_project.advanced_options
        assert "offset" not in get_project.advanced_options
        assert "exposure" not in get_project.advanced_options
        assert get_project.advanced_options["allowed_pairwise_interaction_groups"] == [
            ["A", "B", "C"],
            ["C", "D"],
        ]

        assert isinstance(get_project.created, datetime.datetime)

        assert get_project.partition["cv_method"] == "random"

    @responses.activate
    def test_get_monotonicity_project(self):
        mono_config = {
            "onlyIncludeMonotonicBlueprints": True,
            "defaultMonotonicDecreasingFeaturelistId": "5ae04c0d962d7410683073cb",
            "defaultMonotonicIncreasingFeaturelistId": "5ae04c0f962d7410683073cc",
        }
        project_response = construct_dummy_aimed_project()
        project_response["advancedOptions"].update(**mono_config)
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/",
            body=json.dumps(project_response),
            status=200,
            content_type="application/json",
        )

        get_project = Project.get("p-id")

        assert (
            get_project.advanced_options["default_monotonic_increasing_featurelist_id"]
            == mono_config["defaultMonotonicIncreasingFeaturelistId"]
        )
        assert (
            get_project.advanced_options["default_monotonic_decreasing_featurelist_id"]
            == mono_config["defaultMonotonicDecreasingFeaturelistId"]
        )
        assert (
            get_project.advanced_options["only_include_monotonic_blueprints"]
            == mono_config["onlyIncludeMonotonicBlueprints"]
        )

    @responses.activate
    def test_get_project_weights_offset_exposure(self):
        """
        Test get project with a weights, offset and exposure columns
        """
        # given a project with a weights column
        pid = "582e4309100d2b13b939b223"
        weight_var = "weight_variable"
        offset_var = ["offset_variable"]
        exposure_var = "exposure_variable"
        events_count_var = "events_count"
        project_data = construct_dummy_aimed_project()
        project_data["advancedOptions"]["weights"] = weight_var
        project_data["advancedOptions"]["offset"] = offset_var
        project_data["advancedOptions"]["exposure"] = exposure_var
        project_data["advancedOptions"]["eventsCount"] = events_count_var
        # and a mocked API GET /projects/<pid>/ response
        responses.add(
            responses.GET,
            "https://host_name.com/projects/{}/".format(pid),
            body=json.dumps(project_data),
            status=200,
            content_type="application/json",
        )

        # when method Project.get is called
        project = Project.get(pid)

        # then the resulting object contains weights in advanced options
        assert project.advanced_options["weights"] == weight_var
        assert project.advanced_options["offset"] == offset_var
        assert project.advanced_options["exposure"] == exposure_var
        assert project.advanced_options["events_count"] == events_count_var

    @responses.activate
    def test_get_feature_discovery_project(self):
        pid = "582e4309100d2b13b939b223"
        project_data = construct_dummy_aimed_project()
        project_data["useFeatureDiscovery"] = True
        responses.add(
            responses.GET,
            "https://host_name.com/projects/{}/".format(pid),
            body=json.dumps(project_data),
            status=200,
            content_type="application/json",
        )

        # when method Project.get is called
        project = Project.get(pid)
        assert project.use_feature_discovery is True

    @responses.activate
    def test_delete_project(self):
        """
        Test delete project
        """
        responses.add(responses.DELETE, "https://host_name.com/projects/p-id/", status=204)

        project = Project("p-id")
        del_result = project.delete()
        self.assertEqual(responses.calls[0].request.method, "DELETE")
        self.assertIsNone(del_result)

    @responses.activate
    def test_create_non_ascii_filename(self):
        prep_successful_project_creation_responses()
        # tests/fixtures/日本/データ.csv
        path = fixture_file_path(u"日本/データ.csv")
        Project.create(path)

        assert responses.calls[0].request.url == "https://host_name.com/projects/"
        request_message = responses.calls[0].request.body
        with open(path, "rb") as fd:
            assert fd.read() in request_message.to_string()

    @responses.activate
    def test_create_non_ascii_filepath_is_okay(self):
        """As long as the file name is ascii, we're good"""
        prep_successful_project_creation_responses()
        path = fixture_file_path(u"日本/synthetic-100.csv")

        Project.create(path)
        # decoded to str implicitly
        self.assertEqual(responses.calls[0].request.url, "https://host_name.com/projects/")
        request_message = responses.calls[0].request.body
        with open(path, "rb") as fd:
            self.assertIn(fd.read(), request_message.to_string())

    @responses.activate
    def test_clone(self):
        source_pid = prep_successful_project_cloning_responses()
        project = Project.get(source_pid)
        new_project = project.clone_project()
        assert new_project.id != project.id
        assert new_project.project_name == "Copy of {}".format(project.project_name)
        assert responses.calls[0].request.url == "https://host_name.com/projects/{}/".format(
            source_pid
        )
        assert responses.calls[1].request.url == "https://host_name.com/projectClones/"
        assert responses.calls[1].request.body == json.dumps({"projectId": source_pid}).encode(
            "utf-8"
        )
        assert responses.calls[-1].request.url == "https://host_name.com/projects/{}/".format(
            new_project.id
        )

    @responses.activate
    def test_clone_with_name(self):
        new_name = "this is my next project"
        source_pid = prep_successful_project_cloning_responses(new_name)
        project = Project.get(source_pid)
        new_project = project.clone_project(new_name)
        assert new_project.id != project.id
        assert new_project.project_name == new_name
        assert responses.calls[0].request.url == "https://host_name.com/projects/{}/".format(
            source_pid
        )
        assert responses.calls[1].request.url == "https://host_name.com/projectClones/"
        assert responses.calls[1].request.body == json.dumps(
            {"projectId": source_pid, "projectName": new_name}
        ).encode("utf-8")
        assert responses.calls[-1].request.url == "https://host_name.com/projects/{}/".format(
            new_project.id
        )

    @responses.activate
    def test_rename(self):
        responses.add(
            responses.PATCH,
            "https://host_name.com/projects/p-id/",
            body="",
            status=200,
            content_type="application/json",
        )
        project = Project("p-id")
        project.rename("new name")
        assert responses.calls[0].request.method == "PATCH"
        payload = request_body_to_json(responses.calls[0].request)
        assert project.project_name == payload["projectName"] == "new name"

    @responses.activate
    def test_unlock_holdout(self):
        responses.add(
            responses.PATCH,
            "https://host_name.com/projects/p-id/",
            body="",
            status=200,
            content_type="application/json",
        )
        project = Project("p-id")
        upd_data = project.unlock_holdout()
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        payload = request_body_to_json(responses.calls[0].request)
        self.assertTrue(payload["holdoutUnlocked"])
        self.assertTrue(upd_data)

    @responses.activate
    def test_set_worker_count(self):
        responses.add(
            responses.PATCH,
            "https://host_name.com/projects/p-id/",
            body="",
            status=200,
            content_type="application/json",
        )
        project = Project("p-id")
        project.set_worker_count(20)
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        payload = request_body_to_json(responses.calls[0].request)
        self.assertTrue(payload["workerCount"])

    @responses.activate
    def test_project_create_uses_read_timeout(self):
        """
        Check that the read_timeout parameter gets passed correctly down into
        the `requests` library

        The 400 is just because it's a simple way to short-circuit out of the project
        creation flow, which uses 3 total requests.
        """
        path = fixture_file_path("synthetic-100.csv")

        responses.add(responses.POST, "https://host_name.com/projects/", body="", status=400)
        with mock.patch.object(
            Project._client, "request", wraps=Project._client.request
        ) as mock_request:
            try:
                Project.create(path, read_timeout=2)
            except errors.ClientError:
                pass

            timeout = mock_request.call_args[1]["timeout"]
            assert timeout[0] == 6.05  # Connect timeout; specified in RESTClientObject
            assert timeout[1] == 2  # Read timeout; specified in the function call

    @responses.activate
    def test_by_upload_file_path(self):
        """
        Project.create(
            'synthetic-100.csv')
        """
        prep_successful_project_creation_responses()
        path = fixture_file_path("synthetic-100.csv")

        Project.create(path)
        # decoded to str implicitly
        self.assertEqual(responses.calls[0].request.url, "https://host_name.com/projects/")
        request_message = responses.calls[0].request.body
        with open(path, "rb") as fd:
            self.assertIn(fd.read(), request_message.to_string())

    def test_by_upload_file_path_fail(self):
        """
        Bad file name: Project.create('meh.csv')
        """
        path = fixture_file_path("does_not_exist.csv")
        with self.assertRaises(errors.InputNotUnderstoodError):
            Project.create(path)

    @responses.activate
    def test_by_upload_file_content(self):
        """
        Project.create(b'lalalala')
        """
        prep_successful_project_creation_responses()

        content_line = six.b("one,two\n" + "1,2" * 100)
        Project.create(content_line)
        # decoded to str implicitly
        request_message = responses.calls[0].request.body
        self.assertIn(content_line, request_message.to_string())

    @responses.activate
    def test_project_create_filename(self):
        """
        Project.create(b'lalalala', dataset_filename='my test dataset')
        """
        prep_successful_project_creation_responses()

        content_line = six.b("one,two\n" + "1,2" * 100)
        dataset_filename = "my test dataset"
        Project.create(content_line, dataset_filename=dataset_filename)

        fields = responses.calls[0].request.body.fields
        # Test that the provided dataset_filename was used in the request to api
        assert "file" in fields
        assert fields["file"][0] == dataset_filename

    @responses.activate
    def test_by_upload_file_path_pathlib(self):
        """
        Project.create(Path('synthetic-100.csv'))
        We will use `DummyPathObject` class as a sample of pathlib.Path object,
        as pathlib.Path is not supported for python lower than 3.4 without having extra dependency.
        """

        class DummyPathObject:
            def __init__(self, path):
                self.path = path

            def __str__(self):
                return str(self.path)

        prep_successful_project_creation_responses()
        path_str = fixture_file_path("synthetic-100.csv")
        path_obj = DummyPathObject(path_str)
        Project.create(path_obj)
        # decoded to str implicitly
        self.assertEqual(responses.calls[0].request.url, "https://host_name.com/projects/")
        request_message = responses.calls[0].request.body
        with open(path_str, "rb") as fd:
            self.assertIn(fd.read(), request_message.to_string())

    @responses.activate
    def test_by_upload_file_path_pathlib_invalid(self):
        """
        Bad file name: Project.create(Path('meh.csv'))
        We will use `DummyPathObject` class as a sample of pathlib.Path object,
        as pathlib.Path is not supported for python lower than 3.4 without having extra dependency.
        """

        class DummyPathObject:
            def __init__(self, path):
                self.path = path

            def __str__(self):
                return str(self.path)

        path_str = fixture_file_path("does_not_exist.csv")
        path_obj = DummyPathObject(path_str)
        with self.assertRaises(errors.InputNotUnderstoodError):
            Project.create(path_obj)

    def test_by_upload_content_encoding(self):
        """
        Bad content encoding Project.create(u'lalalala')
        """
        content_line = u"lalalala"
        with self.assertRaises(errors.InputNotUnderstoodError):
            Project.create(content_line)

    @responses.activate
    def test_by_upload_from_fd(self):
        """
        Project.create(
          sourcedata=open('synthetic-100.csv'))
        """
        prep_successful_project_creation_responses()

        path = fixture_file_path("synthetic-100.csv")

        with open(path, "rb") as fd:
            Project.create(sourcedata=fd)
            request_message = responses.calls[0].request.body

            with open(path, "rb") as fd2:
                self.assertIn(fd2.read(), request_message.to_string())

        file_like = six.StringIO("thra\ntata\nrata")
        Project.create(sourcedata=file_like)

        # decoded to str implicitly
        request_message = responses.calls[3].request.body
        self.assertIn(six.b("thra\ntata\nrata"), request_message.to_string())

    @responses.activate
    def test_by_upload_file_seek(self):
        """
        Seek to EOF Project.create(
            open('synthetic-100.csv')
        """
        prep_successful_project_creation_responses()

        path = fixture_file_path("synthetic-100.csv")
        with open(path, "rb") as fd:
            fd.seek(20000000)
            Project.create(fd)
            # decoded to str implicitly

            request_message = responses.calls[0].request.body
            with open(path, "rb") as fd2:
                self.assertIn(fd2.read(), request_message.to_string())

    @responses.activate
    def test_by_upload_file_closed(self):
        """
        Closed fd Project.create(
            open('synthetic-100.csv')
        """
        path = fixture_file_path("synthetic-100.csv")
        responses.add(
            responses.POST, "https://host_name.com/projects/p-id/upload/", body="", status=200,
        )
        fd = open(path)
        fd.close()
        with self.assertRaises(ValueError):
            Project.create(fd)

    @responses.activate
    def test_upload_by_file_url(self):
        """
        Project.create('http:/google.com/datarobot.csv')
        """
        prep_successful_project_creation_responses()

        link = "http:/google.com/datarobot.csv"
        Project.create(sourcedata=link)
        request_message = request_body_to_json(responses.calls[0].request)
        assert request_message == {
            "url": link,
            "projectName": "Untitled Project",
        }
        self.assertEqual(responses.calls[0].request.url, "https://host_name.com/projects/")

    @responses.activate
    def test_set_target(self):
        """
        Project('p-id').set_target('SalePrice', metric='RMSE')
        """
        prep_successful_aim_responses()
        opts = AdvancedOptions(
            weights="WeightName",
            offset=["OffsetName1", "OffsetName2"],
            events_count="events_count_col",
            exposure="ExposureName",
            allowed_pairwise_interaction_groups=[["A", "B", "C"], ["C", "D"]],
        )
        upd_project = Project("p-id").set_target(
            "SalePrice", mode=AUTOPILOT_MODE.FULL_AUTO, metric="RMSE", advanced_options=opts
        )

        request_message = request_body_to_json(responses.calls[0].request)
        assert request_message == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "metric": "RMSE",
            "weights": "WeightName",
            "offset": ["OffsetName1", "OffsetName2"],
            "exposure": "ExposureName",
            "eventsCount": "events_count_col",
            "smartDownsampled": False,
            "allowedPairwiseInteractionGroups": [["A", "B", "C"], ["C", "D"]],
            "quickrun": False,
            "autopilotWithFeatureDiscovery": False,
            "featureDiscoverySupervisedFeatureReduction": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(upd_project, Project)
        advanced = upd_project.advanced_options
        assert advanced["response_cap"] is False
        assert advanced["smart_downsampled"] is False

    @responses.activate
    def test_set_target_price(self):
        """
        Project('p-id').set_target('SalePrice', metric='RMSE')
        """
        prep_successful_aim_responses()
        upd_project = Project("p-id").set_target("SalePrice")

        request_message = request_body_to_json(responses.calls[0].request)
        assert request_message == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "quickrun": True,
        }
        self.assertEqual(upd_project.metric, "Gamma Deviance")

    @responses.activate
    def test_set_target_quickrun_mode(self):
        prep_successful_aim_responses()
        Project("p-id").set_target("SalePrice", mode=AUTOPILOT_MODE.QUICK)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["mode"] == AUTOPILOT_MODE.FULL_AUTO
        assert request_body["quickrun"] is True

    @responses.activate
    def test_set_target_with_valid_target_type_binary(self):
        valid_target_type = TARGET_TYPE.BINARY
        prep_successful_aim_responses()
        Project("p-id").set_target("SalePrice", target_type=valid_target_type)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["targetType"] == valid_target_type

    @responses.activate
    def test_set_target_with_valid_target_type_regression(self):
        valid_target_type = TARGET_TYPE.REGRESSION
        prep_successful_aim_responses()
        Project("p-id").set_target("SalePrice", target_type=valid_target_type)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["targetType"] == valid_target_type

    @responses.activate
    def test_set_target_with_valid_target_type_multiclass(self):
        valid_target_type = TARGET_TYPE.MULTICLASS
        prep_successful_aim_responses()
        Project("p-id").set_target("SalePrice", target_type=valid_target_type)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["targetType"] == valid_target_type

    @responses.activate
    def test_set_target_with_valid_target_type_multilabel(self):
        valid_target_type = TARGET_TYPE.MULTILABEL
        prep_successful_aim_responses()
        Project("p-id").set_target("SalePrice", target_type=valid_target_type)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["targetType"] == valid_target_type

    @responses.activate
    def test_set_target_with_invalid_target_type(self):
        invalid_target_type = "blackbox_magic"
        prep_successful_aim_responses()

        with pytest.raises(TypeError) as exc_info:
            Project("p-id").set_target("SalePrice", target_type=invalid_target_type)
        assert_raised_regex(exc_info, "is not a valid target_type")

        # Make sure that we did not actually make an invalid API call
        assert len(responses.calls) == 0

    @responses.activate
    def test_set_target_async_error(self):
        """
        Project('p-id').set_target('SalePrice', metric='RMSE')
        """
        responses.add(
            responses.PATCH,
            "https://host_name.com/projects/p-id/aim/",
            body="",
            status=202,
            adding_headers={"Location": "https://host_name.com/projects/p-id/"},
            content_type="application/json",
        )

        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/",
            body=json.dumps({"status": "ERROR"}),
            status=200,
            content_type="application/json",
        )
        with self.assertRaises(errors.AsyncProcessUnsuccessfulError):
            Project("p-id").set_target(
                "SalePrice", metric="RMSE",
            )
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")

    def test_advanced_options_must_be_object(self):
        with self.assertRaises(TypeError):
            Project("p-id").set_target("SalePrice", advanced_options={"garbage": "in"})

    @responses.activate
    def test_set_blueprint_threshold(self):
        """
        Set blueprint threshold
        """
        prep_successful_aim_responses()

        opts = AdvancedOptions(blueprint_threshold=2)
        upd_project = Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "autopilotWithFeatureDiscovery": False,
            "featureDiscoverySupervisedFeatureReduction": True,
            "blueprintThreshold": 2,
            "smartDownsampled": False,
            "quickrun": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(upd_project, Project)

    @responses.activate
    def test_blend_best_models_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(blend_best_models=True)
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["blendBestModels"] is True

    @responses.activate
    def test_set_blend_best_models_unset_not_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions()
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert "blendBestModels" not in request_body

    @responses.activate
    def test_scoring_code_only_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(scoring_code_only=True)
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["scoringCodeOnly"] is True

    @responses.activate
    def test_set_scoring_code_only_unset_not_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions()
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert "scoringCodeOnly" not in request_body

    @responses.activate
    def test_set_shap_only_mode_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(shap_only_mode=True)
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["shapOnlyMode"] is True

    @responses.activate
    def test_set_shap_only_mode_unset_not_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions()
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert "shapOnlyMode" not in request_body

    @responses.activate
    def test_prepare_model_for_deployment_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(prepare_model_for_deployment=True)
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["prepareModelForDeployment"] is True

    @responses.activate
    def test_set_prepare_model_for_deployment_unset_not_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions()
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert "prepareModelForDeployment" not in request_body

    @responses.activate
    def test_consider_blenders_in_recommendation_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(consider_blenders_in_recommendation=True)
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["considerBlendersInRecommendation"] is True

    @responses.activate
    def test_set_consider_blenders_in_recommendation_unset_not_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions()
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert "considerBlendersInRecommendation" not in request_body

    @responses.activate
    def test_min_secondary_validation_model_count_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(min_secondary_validation_model_count=5)
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["minSecondaryValidationModelCount"] == 5

    @responses.activate
    def test_set_min_secondary_validation_model_count_unset_not_in_payload(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions()
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert "prepareModelForDeployment" not in request_body

    @responses.activate
    def test_set_response_cap(self):
        """
        Set Response Cap
        """
        prep_successful_aim_responses()

        opts = AdvancedOptions(response_cap=0.7)
        upd_project = Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "responseCap": 0.7,
            "smartDownsampled": False,
            "quickrun": True,
            "autopilotWithFeatureDiscovery": False,
            "featureDiscoverySupervisedFeatureReduction": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(upd_project, Project)

    @responses.activate
    def test_set_seed(self):
        """
        Set Response Cap
        """
        prep_successful_aim_responses()

        opts = AdvancedOptions(seed=22)
        upd_project = Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "seed": 22,
            "smartDownsampled": False,
            "quickrun": True,
            "autopilotWithFeatureDiscovery": False,
            "featureDiscoverySupervisedFeatureReduction": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(upd_project, Project)

    @responses.activate
    def test_set_smart_sampled(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(smart_downsampled=True, majority_downsampling_rate=50.5)
        upd_project = Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "smartDownsampled": True,
            "majorityDownsamplingRate": 50.5,
            "quickrun": True,
            "autopilotWithFeatureDiscovery": False,
            "featureDiscoverySupervisedFeatureReduction": True,
        }
        assert isinstance(upd_project, Project)

    @responses.activate
    def test_set_datetime_autopilot_data_sampling_method(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(autopilot_data_sampling_method="latest")
        upd_project = Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "smartDownsampled": False,
            "autopilotWithFeatureDiscovery": False,
            "featureDiscoverySupervisedFeatureReduction": True,
            "autopilotDataSamplingMethod": "latest",
            "quickrun": True,
        }
        assert isinstance(upd_project, Project)

    @responses.activate
    def test_set_run_leakage_removed_feature_list(self):
        """
        Set project with run leakage removed feature list option
        """
        prep_successful_aim_responses()

        opts = AdvancedOptions(run_leakage_removed_feature_list=False)
        Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["runLeakageRemovedFeatureList"] is False

    @responses.activate
    def test_set_autopilot_with_feature_discovery(self):
        """ Set project to run feature discovery
        """
        prep_successful_aim_responses()

        opts = AdvancedOptions(autopilot_with_feature_discovery=True)
        upd_project = Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "autopilotWithFeatureDiscovery": True,
            "featureDiscoverySupervisedFeatureReduction": True,
            "smartDownsampled": False,
            "quickrun": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(upd_project, Project)

    @responses.activate
    def test_set_feature_discovery_supervised_feature_reduction(self):
        """ Set project to run feature discovery superwised feature reduction
        """
        prep_successful_aim_responses()

        opts = AdvancedOptions(feature_discovery_supervised_feature_reduction=True)
        upd_project = Project("p-id").set_target("SalePrice", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "autopilotWithFeatureDiscovery": False,
            "featureDiscoverySupervisedFeatureReduction": True,
            "smartDownsampled": False,
            "quickrun": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(upd_project, Project)

    @responses.activate
    def test_set_target_advance_partition_method_cv(self):
        """
        Set project with advanced partition method
        """
        prep_successful_aim_responses()

        part_method = UserCV(user_partition_col="NumPartitions", cv_holdout_level=1, seed=42)
        p = Project("p-id").set_target("SalePrice", partitioning_method=part_method)
        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "userPartitionCol": "NumPartitions",
            "cvHoldoutLevel": 1,
            "seed": 42,
            "validationType": "CV",
            "cvMethod": "user",
            "quickrun": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(p, Project)

    @responses.activate
    def test_set_target_advance_partition_method_tvh(self):
        """
        Set project with advanced partition method
        """
        prep_successful_aim_responses()
        part_method = UserTVH(
            user_partition_col="NumPartitions",
            validation_level=1,
            training_level=2,
            holdout_level=3,
            seed=42,
        )
        p = Project("p-id").set_target("SalePrice", partitioning_method=part_method)

        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "userPartitionCol": "NumPartitions",
            "holdoutLevel": 3,
            "seed": 42,
            "validationLevel": 1,
            "trainingLevel": 2,
            "validationType": "TVH",
            "cvMethod": "user",
            "quickrun": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(p, Project)

    @responses.activate
    def test_set_target_specify_positive_class(self):
        """
        Set project with advanced partition method
        """
        prep_successful_aim_responses()

        p = Project("p-id").set_target("Forks", positive_class="None or Unspecified")
        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "Forks",
            "positiveClass": "None or Unspecified",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "quickrun": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(p, Project)

    @responses.activate
    @pytest.mark.usefixtures("known_warning")
    def test_set_target_quickrun_param(self):
        """
        Set project with quickrun option
        """
        prep_successful_aim_responses()

        p = Project("p-id").set_target("Forks", mode=AUTOPILOT_MODE.MANUAL, quickrun=True)
        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "Forks",
            "quickrun": True,
            "mode": AUTOPILOT_MODE.FULL_AUTO,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(p, Project)

    @responses.activate
    def test_set_target_comprehensive(self):
        """
        Set project in comprehensive Autopilot
        """
        prep_successful_aim_responses()

        p = Project("p-id").set_target("Forks", mode=AUTOPILOT_MODE.COMPREHENSIVE)
        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {"target": "Forks", "mode": AUTOPILOT_MODE.COMPREHENSIVE}
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(p, Project)

    def test_pass_advance_part_wrong(self):
        with self.assertRaises(TypeError):
            Project("p-id").set_target("SalePrice", partitioning_method={"CV": "TVH"})

    @responses.activate
    def test_set_target_relationships_configuration_id_n_credentials(self):
        relationships_configuration_id = "config-id"
        credentials = [
            {"user": "abc", "password": "qwerty", "catalog_version_id": "v-id", "url": "abc.jsbc"},
            {"credential_id": "cd_r", "catalog_version_id": "v-id", "url": "erf.sql"},
        ]
        project = {
            "id": "p-id",
            "target": "SalePrice",
            "fileName": "Untitled Project.csv",
            "stage": "modeling",
            "relationshipsConfigurationId": relationships_configuration_id,
            "credentials": credentials,
        }
        prep_successful_aim_responses(project=project)

        project = Project("p-id").set_target(
            target="SalePrice",
            relationships_configuration_id=relationships_configuration_id,
            credentials=credentials,
        )

        request_message = request_body_to_json(responses.calls[0].request)
        assert request_message == {
            "target": "SalePrice",
            "mode": AUTOPILOT_MODE.FULL_AUTO,
            "quickrun": True,
            "relationshipsConfigurationId": "config-id",
            "credentials": [
                {
                    "user": "abc",
                    "password": "qwerty",
                    "catalogVersionId": "v-id",
                    "url": "abc.jsbc",
                },
                {"credentialId": "cd_r", "catalogVersionId": "v-id", "url": "erf.sql"},
            ],
        }
        self.assertIsInstance(project, Project)
        self.assertEqual(project.relationships_configuration_id, relationships_configuration_id)
        self.assertEqual(project.credentials, credentials)

    @responses.activate
    def test_pause_autopilot(self):
        """
        Project('p-id').pause_autopilot()
        """
        responses.add(
            responses.POST,
            "https://host_name.com/projects/p-id/autopilot/",
            body="",
            status=202,
            content_type="application/json",
        )
        self.assertTrue(Project("p-id").pause_autopilot())
        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {"command": "stop"}

    @responses.activate
    def test_unpause_autopilot(self):
        """
        Project('p-id').unpause_autopilot()
        """
        responses.add(
            responses.POST,
            "https://host_name.com/projects/p-id/autopilot/",
            body="",
            status=202,
            content_type="application/json",
        )
        self.assertTrue(Project("p-id").unpause_autopilot())
        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {"command": "start"}

    @responses.activate
    def test_get_featurelists(self):
        """project.get_featurelists()

        """
        some_featurelists = [
            {
                "id": "f-id-1",
                "projectId": "p-id",
                "name": "Raw Features",
                "features": ["One Fish", "Two Fish", "Red Fish", "Blue Fish"],
            },
            {
                "id": "f-id-2",
                "projectId": "p-id",
                "name": "Informative Features",
                "features": ["One Fish", "Red Fish", "Blue Fish"],
            },
            {
                "id": "f-id-3",
                "projectId": "p-id",
                "name": "Custom Features",
                "features": ["One Fish", "Blue Fish"],
            },
        ]

        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/featurelists/",
            body=json.dumps(some_featurelists),
            status=200,
            content_type="application/json",
        )
        flists = Project("p-id").get_featurelists()
        for flist in flists:
            self.assertIsInstance(flist, Featurelist)

    def test_create_featurelist_duplicate_features(self):
        project = Project("p-id")
        with pytest.raises(errors.DuplicateFeaturesError) as exc_info:
            project.create_featurelist("test", ["feature", "feature"])
        assert_raised_regex(exc_info, "Can't create featurelist with duplicate features")

    @responses.activate
    def test_start_project(self):
        """
        Project.start("test_name",
                      "file.csv",
                      "a_target",
                      worker_count=4,
                      metric="a_metric")
        """
        path = fixture_file_path("synthetic-100.csv")
        prep_successful_project_creation_responses()

        responses.add(
            responses.PATCH,
            "https://host_name.com/projects/p-id/",
            body="",
            status=200,
            content_type="application/json",
        )

        prep_successful_aim_responses()

        proj = Project.start(
            project_name="test_name",
            sourcedata=path,
            target="a_target",
            worker_count=4,
            metric="a_metric",
        )
        self.assert_project_start_call_order()
        self.assertIsInstance(proj, Project)

    @responses.activate
    def test_start_project_from_dataframe(self):
        """
        Project.start("test_name",
                      "file.csv",
                      "a_target",
                      worker_count=4,
                      metric="a_metric")
        """
        prep_successful_project_creation_responses()

        responses.add(
            responses.PATCH,
            "https://host_name.com/projects/p-id/",
            body="",
            status=200,
            content_type="application/json",
        )

        prep_successful_aim_responses()

        dataframe = pd.DataFrame({"a_target": range(100), "predictor": range(100, 200)})
        proj = Project.start(dataframe, "a_target", "test_name", worker_count=4, metric="a_metric")
        self.assert_project_start_call_order()
        self.assertIsInstance(proj, Project)

    @responses.activate
    def test_start_project_in_manual_mode(self):
        """
        Project.start("test_name",
                      "file.csv",
                      "a_target",
                      worker_count=4,
                      metric="a_metric",
                      autopilot_on=False)
        """
        path = fixture_file_path("synthetic-100.csv")
        prep_successful_project_creation_responses()
        responses.add(
            responses.PATCH,
            "https://host_name.com/projects/p-id/",
            body="",
            status=200,
            content_type="application/json",
        )
        prep_successful_aim_responses()

        proj = Project.start(
            path, "test_name", "a_target", worker_count=4, metric="a_metric", autopilot_on=False
        )
        self.assert_project_start_call_order()
        self.assertIsInstance(proj, Project)

        request_body = request_body_to_json(responses.calls[4].request)
        self.assertEqual(request_body["mode"], AUTOPILOT_MODE.MANUAL)

    @responses.activate
    def test_start_feature_discovery_project(self):
        path = fixture_file_path("synthetic-100.csv")
        prep_successful_project_creation_responses()

        responses.add(
            responses.PATCH,
            "https://host_name.com/projects/p-id/",
            body="",
            status=200,
            content_type="application/json",
        )

        prep_successful_aim_responses()

        proj = Project.start(
            project_name="test_name",
            sourcedata=path,
            target="a_target",
            worker_count=4,
            metric="a_metric",
            relationships_configuration_id="r-id",
        )
        self.assert_project_start_call_order()
        self.assertIsInstance(proj, Project)

        request_body = request_body_to_json(responses.calls[4].request)
        self.assertEqual(request_body["relationshipsConfigurationId"], "r-id")

    @responses.activate
    def test_set_scaleout_mode(self):
        """
        Project('p-id').set_target('SalePrice', metric='RMSE')
        """
        prep_successful_aim_responses()
        opts = AdvancedOptions(scaleout_modeling_mode=SCALEOUT_MODELING_MODE.AUTOPILOT)
        upd_project = Project("p-id").set_target("SalePrice", metric="RMSE", advanced_options=opts)

        request_body = request_body_to_json(responses.calls[0].request)
        self.assertEqual(
            request_body,
            {
                "target": "SalePrice",
                "mode": AUTOPILOT_MODE.FULL_AUTO,
                "metric": "RMSE",
                "smartDownsampled": False,
                "scaleoutModelingMode": SCALEOUT_MODELING_MODE.AUTOPILOT,
                "quickrun": True,
                "autopilotWithFeatureDiscovery": False,
                "featureDiscoverySupervisedFeatureReduction": True,
            },
        )
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(upd_project, Project)

    def assert_project_start_call_order(self):
        """Run this assertion to assert that the expected calls for project.start
        happened, and in the correct order

        This is a terribly brittle test. If we can come up with something
        better, let's go with that.
        """
        # Create Project
        self.assertEqual(responses.calls[0].request.url, "https://host_name.com/projects/")
        self.assertEqual(responses.calls[0].request.method, "POST")

        # Get Project Creation Status
        self.assertEqual(responses.calls[1].request.url, "https://host_name.com/status/status-id/")
        self.assertEqual(responses.calls[1].request.method, "GET")

        # Get Created Project
        self.assertEqual(responses.calls[2].request.url, "https://host_name.com/projects/p-id/")
        self.assertEqual(responses.calls[2].request.method, "GET")

        # Set Worker Count
        self.assertEqual(responses.calls[3].request.url, "https://host_name.com/projects/p-id/")
        self.assertEqual(responses.calls[3].request.method, "PATCH")

        # Set Target
        self.assertEqual(responses.calls[4].request.url, "https://host_name.com/projects/p-id/aim/")
        self.assertEqual(responses.calls[4].request.method, "PATCH")

        # Get Status of Finalizing Project
        self.assertEqual(
            responses.calls[5].request.url, "https://host_name.com/status/some-status/"
        )
        self.assertEqual(responses.calls[5].request.method, "GET")

        # Get Finalized Project
        self.assertEqual(responses.calls[6].request.url, "https://host_name.com/projects/p-id/")
        self.assertEqual(responses.calls[6].request.method, "GET")

    @responses.activate
    def test_set_target_featurelist(self):
        """
        proj.set_target('test_target', featurelist_id=...)
        """
        prep_successful_aim_responses()
        Project("p-id").set_target("a_target", featurelist_id="f-id-3")
        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body["featurelistId"] == "f-id-3"

    @responses.activate
    def test_set_target_monotonicity(self):
        prep_successful_aim_responses()

        opts = AdvancedOptions(
            monotonic_increasing_featurelist_id="fi-id",
            monotonic_decreasing_featurelist_id="fd-id",
            only_include_monotonic_blueprints=True,
        )
        p = Project("p-id").set_target(
            "Forks", positive_class="None or Unspecified", advanced_options=opts
        )
        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "target": "Forks",
            "mode": "auto",
            "smartDownsampled": False,
            "monotonicIncreasingFeaturelistId": "fi-id",
            "monotonicDecreasingFeaturelistId": "fd-id",
            "onlyIncludeMonotonicBlueprints": True,
            "positiveClass": "None or Unspecified",
            "quickrun": True,
            "autopilotWithFeatureDiscovery": False,
            "featureDiscoverySupervisedFeatureReduction": True,
        }
        self.assertEqual(responses.calls[0].request.method, "PATCH")
        self.assertEqual(responses.calls[1].request.method, "GET")
        self.assertIsInstance(p, Project)

    @responses.activate
    def test_attach_file_with_link_goes_to_url(self):
        """
        Project.create('https://google.com/file.csv')
        """
        prep_successful_project_creation_responses()

        link = "http:/google.com/datarobot.csv"
        Project.create(link)
        request_body = request_body_to_json(responses.calls[0].request)
        assert request_body == {
            "url": link,
            "projectName": "Untitled Project",
        }
        self.assertEqual(responses.calls[0].request.url, "https://host_name.com/projects/")

    @responses.activate
    def test_attach_file_with_file_path(self):
        """
        Project.create('synthetic-100.csv')
        """
        prep_successful_project_creation_responses()

        path = fixture_file_path("synthetic-100.csv")
        Project.create(path)

        self.assertEqual(responses.calls[0].request.url, "https://host_name.com/projects/")
        request_message = responses.calls[0].request.body
        with open(path, "rb") as fd:
            self.assertIn(fd.read(), request_message.read())

    @responses.activate
    def test_attach_file_with_non_csv_file_path(self):
        """
        Project.create('dataset.xlsx')
        """
        prep_successful_project_creation_responses()

        path = fixture_file_path("onehundredrows.xlsx")
        Project.create(path)
        self.assertEqual(responses.calls[0].request.url, "https://host_name.com/projects/")
        request_message = responses.calls[0].request.body.read()
        with open(path, "rb") as fd:
            self.assertIn(fd.read(), request_message)
        fname = re.search(b'filename="(.*)"', request_message).group(1)
        self.assertEqual(fname, b"onehundredrows.xlsx")

    @responses.activate
    def test_get_status_underscorizes(self):
        body = json.dumps(
            {"autopilotDone": True, "stage": "modeling", "stageDescription": "Ready for modeling"}
        )
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/status/",
            status=200,
            body=body,
            content_type="application/json",
        )
        status = Project("p-id").get_status()
        self.assertEqual(status["stage"], "modeling")
        self.assertTrue(status["autopilot_done"])

    @responses.activate
    def test_get_blueprints(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/blueprints/",
            status=200,
            body=json.dumps(self.blueprint_api_resp),
            content_type="application/json",
        )
        menu = Project("p-id").get_blueprints()
        for item in menu:
            self.assertIsInstance(item, Blueprint)
            self.assertEqual(item.project_id, "p-id")
        bluepr1 = menu[0]
        bluepr2 = menu[1]
        self.assertIsInstance(bluepr1.processes, list)
        self.assertEqual(bluepr1.processes[0], "Regularized Linear Model Preprocessing v5")
        self.assertEqual(bluepr1.processes[1], "Log Transformed Response")
        self.assertEqual(
            bluepr1.model_type, "Auto-tuned K-Nearest Neighbors Regressor (Minkowski Distance)"
        )
        self.assertIsInstance(bluepr2.processes, list)
        self.assertEqual(bluepr2.processes[0], "Regularized Linear Model Preprocessing v12")
        self.assertEqual(bluepr2.processes[1], "Log Transformed Response")
        self.assertEqual(bluepr2.model_type, "Auto-tuned Stochastic Gradient Descent Regression")

    @responses.activate
    def test_get_monotonicity_blueprints(self):
        mono_config = {
            "supportsMonotonicConstraints": True,
            "monotonicDecreasingFeaturelistId": "5ae04c0d962d7410683073cb",
            "monotonicIncreasingFeaturelistId": "5ae04c0f962d7410683073cc",
        }
        blueprint_resp = [dict(**b) for b in self.blueprint_api_resp]
        [b.update(mono_config) for b in blueprint_resp]
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/blueprints/",
            status=200,
            body=json.dumps(blueprint_resp),
            content_type="application/json",
        )
        blueprints = Project("p-id").get_blueprints()
        for blueprint in blueprints:
            assert (
                blueprint.monotonic_increasing_featurelist_id
                == mono_config["monotonicIncreasingFeaturelistId"]
            )
            assert (
                blueprint.monotonic_decreasing_featurelist_id
                == mono_config["monotonicDecreasingFeaturelistId"]
            )
            assert (
                blueprint.supports_monotonic_constraints
                == mono_config["supportsMonotonicConstraints"]
            )


@responses.activate
def test_create_featurelist(featurelist_server_data, project_id):
    project = Project(project_id)
    name = featurelist_server_data["name"]
    features = featurelist_server_data["features"]

    responses.add(
        responses.POST,
        "https://host_name.com/projects/{}/featurelists/".format(project_id),
        json=featurelist_server_data,
        status=201,
        content_type="application/json",
        adding_headers={
            "Location": "https://host_name.com/projects/{}/featurelists/{}/".format(
                project_id, featurelist_server_data["id"]
            )
        },
    )
    new_flist = project.create_featurelist(name, features)
    assert new_flist.name == name
    assert new_flist.features == features
    assert new_flist.project_id == project_id
    assert new_flist.num_models == featurelist_server_data["numModels"]
    assert new_flist.is_user_created == featurelist_server_data["isUserCreated"]
    assert new_flist.description == featurelist_server_data["description"]
    assert isinstance(new_flist.created, datetime.datetime)


class TestProjectJobListing(SDKTestcase):
    job1_queued_dict = {
        "status": "queue",
        "processes": ["Majority Class Classifier"],
        "projectId": "556902e8100d2b3728d47551",
        "samplePct": 32.0,
        "modelType": "Majority Class Classifier",
        "featurelistId": "556902eb100d2b37d1130771",
        "blueprintId": "89e08076a908e859c07af49bd4aa6a0f",
        "id": "10",
        "isBlocked": False,
        "modelId": "556902ef100d2b37da13077c",
    }
    job2_queued_dict = {
        "status": "queue",
        "processes": ["One-Hot Encoding", "Missing Values Imputed", "RuleFit Classifier"],
        "projectId": "556902e8100d2b3728d47551",
        "samplePct": 32.0,
        "modelType": "RuleFit Classifier",
        "featurelistId": "556902eb100d2b37d1130771",
        "blueprintId": "a8959bc1d46f07fb3dc14db7c1e3fc99",
        "id": "11",
        "isBlocked": False,
        "modelId": "556902ef100d2b37da13077d",
    }
    job3_queued_dict = {
        "status": "queue",
        "processes": ["One-Hot Encoding", "Missing Values Imputed", "RuleFit Classifier"],
        "projectId": "556902e8100d2b3728d47551",
        "samplePct": 64.0,
        "modelType": "RuleFit Classifier",
        "featurelistId": "556902eb100d2b37d1130771",
        "blueprintId": "a8959bc1d46f07fb3dc14db7c1e3fc99",
        "id": "17",
        "isBlocked": False,
        "modelId": "556902ef100d2b37da13077d",
    }
    job1_inprogress_dict = dict(job1_queued_dict, status="inprogress")
    job2_inprogress_dict = dict(job2_queued_dict, status="inprogress")

    predict_job_queued_dict = {
        "status": "queue",
        "projectId": "56b62892ccf94e7e939c89c8",
        "message": "",
        "id": "27",
        "isBlocked": False,
        "modelId": "56b628b7ccf94e0444bb8152",
    }

    predict_job_errored_dict = {
        "status": "error",
        "projectId": "56b62892ccf94e7e939c89c8",
        "message": "",
        "id": "23",
        "isBlocked": False,
        "modelId": "56b628b7ccf94e0444bb8152",
    }

    all_jobs_response = {
        "count": 4,
        "next": None,
        "jobs": [
            {
                "status": "inprogress",
                "url": "https://host_name.com/projects/p-id/modelJobs/1/",
                "id": "1",
                "jobType": "model",
                "isBlocked": False,
                "projectId": "p-id",
            },
            {
                "status": "inprogress",
                "url": "https://host_name.com/projects/p-id/modelJobs/2/",
                "id": "2",
                "jobType": "model",
                "isBlocked": False,
                "projectId": "p-id",
            },
            {
                "status": "queue",
                "url": "https://host_name.com/projects/p-id/predictJobs/3/",
                "id": "3",
                "jobType": "predict",
                "isBlocked": False,
                "projectId": "p-id",
            },
            {
                "status": "queue",
                "url": "https://host_name.com/projects/p-id/modelJobs/4/",
                "id": "4",
                "jobType": "model",
                "isBlocked": False,
                "projectId": "p-id",
            },
        ],
        "previous": None,
    }

    all_error_jobs_response = {
        "count": 1,
        "next": None,
        "jobs": [
            {
                "status": "error",
                "url": "https://host_name.com/projects/p-id/predictJobs/3/",
                "id": "3",
                "jobType": "predict",
                "isBlocked": False,
                "projectId": "p-id",
            }
        ],
        "previous": None,
    }

    @responses.activate
    def test_get_all_jobs(self):
        body = json.dumps(self.all_jobs_response)
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/jobs/",
            status=200,
            body=body,
            content_type="application/json",
        )
        jobs = Project("p-id").get_all_jobs()
        self.assertEqual(len(jobs), 4)

    @responses.activate
    def test_get_all_errored_jobs(self):
        body = json.dumps(self.all_error_jobs_response)
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/jobs/",
            status=200,
            body=body,
            content_type="application/json",
        )
        jobs = Project("p-id").get_all_jobs(status=QUEUE_STATUS.ERROR)
        self.assertEqual(
            responses.calls[0].request.url, "https://host_name.com/projects/p-id/jobs/?status=error"
        )
        self.assertEqual(len(jobs), 1)

    @responses.activate
    def test_get_model_jobs(self):
        job_dict = self.job1_queued_dict
        body = json.dumps([job_dict])
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/modelJobs/",
            status=200,
            body=body,
            content_type="application/json",
        )
        jobs = Project("p-id").get_model_jobs(status=QUEUE_STATUS.QUEUE)
        self.assertEqual(
            responses.calls[0].request.url,
            "https://host_name.com/projects/p-id/modelJobs/?status=queue",
        )
        self.assertEqual(len(jobs), 1)
        received_dict = jobs[0].__dict__
        expected_dict = ModelJob(job_dict).__dict__
        # because two `Model()` instances don't count as equal even if
        # their IDs are equal, we have to compare them separately.
        received_model = received_dict.pop("model")
        expected_model = expected_dict.pop("model")
        assert received_model.id == expected_model.id
        assert received_dict == expected_dict

    @responses.activate
    def test_get_model_jobs_requests_all_by_default(self):
        job_dict = self.job1_queued_dict
        body = json.dumps([job_dict])
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/modelJobs/",
            status=200,
            body=body,
            content_type="application/json",
        )
        jobs = Project("p-id").get_model_jobs()
        self.assertEqual(
            responses.calls[0].request.url, "https://host_name.com/projects/p-id/modelJobs/"
        )
        self.assertEqual(len(jobs), 1)
        received_dict = jobs[0].__dict__
        expected_dict = ModelJob(job_dict).__dict__
        # because two `Model()` instances don't count as equal even if
        # their IDs are equal, we have to compare them separately.
        received_model = received_dict.pop("model")
        expected_model = expected_dict.pop("model")
        assert received_model.id == expected_model.id
        assert received_dict == expected_dict

    @responses.activate
    def test_get_predict_jobs(self):
        body = json.dumps([self.predict_job_errored_dict])
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/predictJobs/",
            status=200,
            body=body,
            content_type="application/json",
        )
        jobs = Project("p-id").get_predict_jobs(status=QUEUE_STATUS.ERROR)
        self.assertEqual(
            responses.calls[0].request.url,
            "https://host_name.com/projects/p-id/predictJobs/?status=error",
        )
        self.assertEqual(len(jobs), 1)
        result_job = jobs[0]
        expected_job = PredictJob(self.predict_job_errored_dict)
        self.assertEqual(expected_job.id, result_job.id)
        self.assertEqual(expected_job.project, result_job.project)

    @responses.activate
    def test_get_predict_jobs_default(self):
        body = json.dumps([self.predict_job_queued_dict])
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/predictJobs/",
            status=200,
            body=body,
            content_type="application/json",
        )
        jobs = Project("p-id").get_predict_jobs()
        self.assertEqual(
            responses.calls[0].request.url, "https://host_name.com/projects/p-id/predictJobs/"
        )
        self.assertEqual(len(jobs), 1)
        result_job = jobs[0]
        expected_job = PredictJob(self.predict_job_queued_dict)
        self.assertEqual(expected_job.id, result_job.id)
        self.assertEqual(expected_job.project, result_job.project)

    @patch("datarobot.Project.get_model_jobs")
    def test_job_status_counter(self, mock_get_jobs):
        project = Project("p-id")
        jobs0 = []
        jobs1 = [ModelJob(self.job1_queued_dict)]
        jobs2 = [ModelJob(self.job1_inprogress_dict), ModelJob(self.job2_queued_dict)]
        jobs3 = [ModelJob(self.job2_inprogress_dict)]
        mock_get_jobs.side_effect = [jobs0, jobs1, jobs2, jobs3]
        assert project._get_job_status_counts() == (0, 0)
        assert project._get_job_status_counts() == (0, 1)
        assert project._get_job_status_counts() == (1, 1)
        assert project._get_job_status_counts() == (1, 0)

    def test_wait_for_autopilot(self):
        def make_status(autopilot_done):
            return {"autopilot_done": autopilot_done}

        project = Project("p-id")
        with mock.patch.object(Project, "_autopilot_status_check") as mock_autopilot_check:
            mock_autopilot_check.side_effect = (
                [make_status(False)] * 3 + [make_status(True)] + [make_status(False)]
            )

            project.wait_for_autopilot(check_interval=0.5, verbosity=0)
            assert mock_autopilot_check.call_count == 4

    def test_wait_for_autopilot_timeout(self):
        project = Project("p-id")
        with mock.patch.object(Project, "_autopilot_status_check") as mock_autopilot_check:
            mock_autopilot_check.return_value = {"autopilot_done": False}
            self.assertRaises(
                errors.AsyncTimeoutError,
                project.wait_for_autopilot,
                check_interval=10,
                verbosity=0,
                timeout=0.2,
            )

    def test_wait_for_autopilot_target_not_set_is_error(self):
        project = Project("p-id")
        with mock.patch.object(Project, "get_status") as mock_get_status:
            mock_get_status.return_value = {
                "autopilot_done": False,
                "stage": PROJECT_STAGE.AIM,
            }
            with pytest.raises(RuntimeError) as exc_info:
                project.wait_for_autopilot(check_interval=10, verbosity=0, timeout=0.2)
            assert_raised_regex(exc_info, "target has not been set")

    def test_wait_for_autopilot_mode_incorrect_is_error(self):
        project = Project("p-id")
        with mock.patch.multiple(
            Project,
            get_status=lambda _self: {"autopilot_done": False, "stage": PROJECT_STAGE.MODELING},
            _server_data=lambda _self, _: construct_dummy_aimed_project(autopilot_mode=2),
        ):
            with pytest.raises(RuntimeError) as exc_info:
                project.wait_for_autopilot(check_interval=10, verbosity=0, timeout=0.2)

            assert_raised_regex(exc_info, "mode is not full auto")


class TestProjectList(SDKTestcase):
    def setUp(self):
        self.raw_return = """
        [
            {
            "project_name": "Api project",
            "_id": "54c627fa100d2b2c7002a489"
            },
            {
            "_id": "54c78125100d2b2fe3b296b6",
            "project_name": "Untitled project"
            }
        ]
        """

    @responses.activate
    def test_list_projects(self):
        """
        Test list projects
        """
        responses.add(
            responses.GET,
            "https://host_name.com/projects/",
            body=self.raw_return,
            status=200,
            content_type="application/json",
        )
        project_lists = Project.list()
        assert isinstance(project_lists, list)
        assert project_lists[0].id == "54c627fa100d2b2c7002a489"
        assert project_lists[0].project_name == "Api project"
        assert project_lists[1].id == "54c78125100d2b2fe3b296b6"
        assert project_lists[1].project_name == "Untitled project"

    @responses.activate
    def test_with_manual_search_params(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/?projectName=x",
            body=self.raw_return,
            status=200,
            content_type="application/json",
            match_querystring=True,
        )
        Project("p-id").list(search_params={"project_name": "x"})
        self.assertEqual(
            responses.calls[0].request.url, "https://host_name.com/projects/?projectName=x"
        )

    @responses.activate
    def test_with_bad_search_params(self):
        with self.assertRaises(TypeError):
            Project("p-id").list(search_params=12)


class TestGetModels(SDKTestcase, URLParamsTestCase):
    def setUp(self):
        super(TestGetModels, self).setUp()
        self.raw_data = """
        [
    {
    "featurelistId": "556cdfbd100d2b10048c7941",
    "processes": ["One", "Two", "Three"],
    "featurelistName": "Informative Features",
    "projectId": "p-id",
    "samplePct": 64,
    "modelType": "AVG Blender",
    "metrics": {
        "AUC": {
            "holdout": 0.76603,
            "validation": 0.64141,
            "crossValidation": 0.7625240000000001
        },
        "Rate@Top5%": {
            "holdout": 1,
            "validation": 0.5,
            "crossValidation": 0.9
        },
        "Rate@TopTenth%": {
            "holdout": 1,
            "validation": 1,
            "crossValidation": 1
        },
        "RMSE": {
            "holdout": 0.42054,
            "validation": 0.44396,
            "crossValidation": 0.40162000000000003
        },
        "LogLoss": {
            "holdout": 0.53707,
            "validation": 0.58051,
            "crossValidation": 0.5054160000000001
        },
        "FVE Binomial": {
            "holdout": 0.17154,
            "validation": 0.03641,
            "crossValidation": 0.17637399999999998
        },
        "Gini Norm": {
            "holdout": 0.53206,
            "validation": 0.28282,
            "crossValidation": 0.525048
        },
        "Rate@Top10%": {
            "holdout": 1,
            "validation": 0.25,
            "crossValidation": 0.7
        }
    },
    "blueprintId": "a4fd9d17a8ca62ee00590dd704dae6a8",
    "id": "556ce973100d2b6e51ca9657"
},
    {
    "featurelistId": "556cdfbd100d2b10048c7941",
    "processes": ["One", "Two", "Three"],
    "featurelistName": "Informative Features",
    "projectId": "p-id",
    "samplePct": 64,
    "modelType": "AVG Blender",
    "metrics": {
        "AUC": {
            "holdout": 0.76603,
            "validation": 0.64141,
            "crossValidation": 0.7625240000000001
        },
        "Rate@Top5%": {
            "holdout": 1,
            "validation": 0.5,
            "crossValidation": 0.9
        },
        "Rate@TopTenth%": {
            "holdout": 1,
            "validation": 1,
            "crossValidation": 1
        },
        "RMSE": {
            "holdout": 0.42054,
            "validation": 0.44396,
            "crossValidation": 0.40162000000000003
        },
        "LogLoss": {
            "holdout": 0.53707,
            "validation": 0.58051,
            "crossValidation": 0.5054160000000001
        },
        "FVE Binomial": {
            "holdout": 0.17154,
            "validation": 0.03641,
            "crossValidation": 0.17637399999999998
        },
        "Gini Norm": {
            "holdout": 0.53206,
            "validation": 0.28282,
            "crossValidation": 0.525048
        },
        "Rate@Top10%": {
            "holdout": 1,
            "validation": 0.25,
            "crossValidation": 0.7
        }
    },
    "blueprintId": "a4fd9d17a8ca62ee00590dd704dae6a8",
    "id": "556ce973100d2b6e51ca9658"
}
]
        """

    @responses.activate
    def test_get_models_ordered_by_metric_by_default(self):
        """
        Project('p-id').get_models()
        """
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/models/?orderBy=-metric",
            body=self.raw_data,
            status=200,
            content_type="application/json",
            match_querystring=True,
        )
        leaderboard = Project("p-id").get_models()
        self.assertNotEqual(0, len(leaderboard))
        for item in leaderboard:
            self.assertIsInstance(item, Model)
        self.assertEqual(leaderboard[0].id, "556ce973100d2b6e51ca9657")
        self.assertEqual(leaderboard[1].id, "556ce973100d2b6e51ca9658")
        self.assertEqual(leaderboard[0].project_id, "p-id")
        self.assertEqual(leaderboard[1].project_id, "p-id")

    @responses.activate
    def test_get_models_ordered_by_specific_field(self):
        # ordering
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/models/?orderBy=samplePct",
            body=self.raw_data,
            status=200,
            content_type="application/json",
            match_querystring=True,
        )
        Project("p-id").get_models(order_by="samplePct")
        self.assertEqual(
            responses.calls[0].request.url,
            "https://host_name.com/projects/p-id/models/?orderBy=samplePct",
        )

    @responses.activate
    def test_get_models_orderd_by_two_fields(self):
        # ordering by two fields
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/models/?orderBy=samplePct%2C-metric",
            body=self.raw_data,
            status=200,
            content_type="application/json",
            match_querystring=True,
        )
        Project("p-id").get_models(order_by=["sample_pct", "-metric"])
        self.assertEqual(
            responses.calls[0].request.url,
            "https://host_name.com/projects/p-id/models/?orderBy=samplePct%2C-metric",
        )

    @responses.activate
    def test_get_models_two_order_fields_and_filtering(self):
        # ordering by two fields plus filtering
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/models/"
            "?withMetric=RMSE&orderBy=metric%2C-samplePct",
            body=self.raw_data,
            status=200,
            content_type="application/json",
            match_querystring=True,
        )
        Project("p-id").get_models(order_by=["metric", "-sample_pct"], with_metric="RMSE")

    @responses.activate
    def test_get_monotonicity_models(self):
        mono_config = {
            "supportsMonotonicConstraints": True,
            "monotonicIncreasingFeaturelistId": "5ae04c0d962d7410683073cb",
            "monotonicDecreasingFeaturelistId": "5ae04c0f962d7410683073cc",
        }
        models_resp = json.loads(self.raw_data)
        [m.update(mono_config) for m in models_resp]
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/models/?withMetric=RMSE&orderBy=-metric",
            body=json.dumps(models_resp),
            status=200,
            content_type="application/json",
            match_querystring=True,
        )

        models = Project("p-id").get_models(with_metric="RMSE")

        for model in models:
            assert (
                model.supports_monotonic_constraints == mono_config["supportsMonotonicConstraints"]
            )
            assert (
                model.monotonic_increasing_featurelist_id
                == mono_config["monotonicIncreasingFeaturelistId"]
            )
            assert (
                model.monotonic_decreasing_featurelist_id
                == mono_config["monotonicDecreasingFeaturelistId"]
            )

    def test_order_by_with_unexpected_param_fails(self):
        with pytest.raises(ValueError) as exc_info:
            Project("p-id").get_models(order_by="someThing")
        assert_raised_regex(exc_info, "Provided order_by attribute")

    def test_order_by_with_bad_param_fails(self):
        with pytest.raises(TypeError) as exc_info:
            Project("p-id").get_models(order_by=True)
        assert_raised_regex(exc_info, "Provided order_by attribute")

    def test_with_bad_search_param_fails(self):
        with self.assertRaises(TypeError):
            Project("p-id").get_models(search_params=True)

    def _canonize_order_by_handles_none(self):
        proj = Project("p-id")
        self.assertIsNone(proj._canonize_order_by(None))

    @responses.activate
    def test_get_models_search_starred(self):
        responses.add(
            responses.GET,
            "https://host_name.com/projects/p-id/models/?orderBy=-metric&isStarred=false",
            body=self.raw_data,
            status=200,
            content_type="application/json",
            match_querystring=True,
        )
        Project("p-id").get_models(search_params={"is_starred": False})
        assert_urls_equal(
            responses.calls[0].request.url,
            "https://host_name.com/projects/p-id/models/?orderBy=-metric&isStarred=false",
        )


@responses.activate
def retrieve_smart_sampled_project(project_url, project_id, smart_sampled_project_server_data):
    responses.add(
        responses.GET,
        project_url,
        body=json.dumps(smart_sampled_project_server_data),
        status=200,
        content_type="application/json",
    )
    smart_sampled_project = Project.get(project_id)
    advanced = smart_sampled_project.advanced_options
    assert advanced["smart_downsampled"] is True
    rate_key = "majority_downsampling_rate"
    assert advanced[rate_key] == smart_sampled_project_server_data[rate_key]


@responses.activate
def test_get_rating_tables(
    rating_table_backend_generated_json,
    rating_table_uploaded_but_not_modeled_json,
    rating_table_uploaded_and_modeled_json,
):

    sample_rating_tables = json.dumps(
        {
            "data": [
                json.loads(rating_table_backend_generated_json),
                json.loads(rating_table_uploaded_but_not_modeled_json),
                json.loads(rating_table_uploaded_and_modeled_json),
            ]
        }
    )

    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/ratingTables/",
        body=sample_rating_tables,
        status=200,
        content_type="application/json",
    )
    rating_tables = Project("p-id").get_rating_tables()

    assert len(rating_tables) == 3
    for rt in rating_tables:
        assert isinstance(rt, RatingTable)
        assert rt.id


@responses.activate
def test_get_rating_tables_no_tables():

    sample_rating_tables = json.dumps({"data": []})

    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/ratingTables/",
        body=sample_rating_tables,
        status=200,
        content_type="application/json",
    )
    rating_tables = Project("p-id").get_rating_tables()

    assert len(rating_tables) == 0


@responses.activate
def test_get_rating_table_models(rating_table_model_server_data):
    sample_rating_table_model = json.dumps(
        [rating_table_model_server_data, rating_table_model_server_data]
    )

    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/ratingTableModels/",
        body=sample_rating_table_model,
        status=200,
        content_type="application/json",
    )
    rating_table_models = Project("p-id").get_rating_table_models()

    assert len(rating_table_models) == 2
    for rtm in rating_table_models:
        assert isinstance(rtm, RatingTableModel)
        assert rtm.id
        assert rtm.rating_table_id


@responses.activate
def test_get_rating_table_models_no_models():
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/ratingTableModels/",
        body="[]",
        status=200,
        content_type="application/json",
    )
    rating_table_models = Project("p-id").get_rating_table_models()

    assert len(rating_table_models) == 0


@responses.activate
@pytest.mark.parametrize("extra_keys", [{}, {"newKey": "blah"}])
def test_get_access_list(project_server_data, project_endpoint, extra_keys):
    project = Project.from_server_data(project_server_data)
    access_list = {
        "data": [
            {
                "username": "me@datarobot.com",
                "userId": "1234deadbeeffeeddead4321",
                "role": "OWNER",
                "canShare": True,
            }
        ],
        "count": 1,
        "previous": None,
        "next": None,
    }
    access_record = access_list["data"][0]
    access_record.update(extra_keys)
    responses.add(
        responses.GET, "{}{}/accessControl/".format(project_endpoint, project.id), json=access_list
    )

    response = project.get_access_list()
    assert len(response) == 1

    share_info = response[0]
    assert share_info.username == access_record["username"]
    assert share_info.user_id == access_record["userId"]
    assert share_info.role == access_record["role"]
    assert share_info.can_share == access_record["canShare"]


@responses.activate
def test_share(project_endpoint, project_server_data):
    project = Project.from_server_data(project_server_data)
    request = SharingAccess("me@datarobot.com", "EDITOR")
    responses.add(
        responses.PATCH, "{}{}/accessControl/".format(project_endpoint, project.id), status=204
    )

    project.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {"data": [{"username": "me@datarobot.com", "role": "EDITOR"}]}


@responses.activate
def test_remove_share(project_endpoint, project_server_data):
    project = Project.from_server_data(project_server_data)
    request = SharingAccess("me@datarobot.com", None)
    responses.add(
        responses.PATCH, "{}{}/accessControl/".format(project_endpoint, project.id), status=204
    )

    project.share([request])

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {"data": [{"username": "me@datarobot.com", "role": None}]}


@responses.activate
def test_share_with_grant(project_endpoint, project_server_data):
    project = Project.from_server_data(project_server_data)
    responses.add(
        responses.PATCH, "{}{}/accessControl/".format(project_endpoint, project.id), status=204
    )

    project.share(
        [
            SharingAccess("me@datarobot.com", "EDITOR", can_share=True),
            SharingAccess("other@datarobot.com", "CONSUMER", can_share=False),
        ]
    )

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {
        "data": [
            {"username": "me@datarobot.com", "role": "EDITOR", "canShare": True},
            {"username": "other@datarobot.com", "role": "CONSUMER", "canShare": False},
        ]
    }


@responses.activate
@pytest.mark.parametrize("send_notification", [True, False])
def test_share_send_notification(send_notification, project_endpoint, project_server_data):
    project = Project.from_server_data(project_server_data)
    request = SharingAccess("me@datarobot.com", "EDITOR")
    responses.add(
        responses.PATCH, "{}{}/accessControl/".format(project_endpoint, project.id), status=204
    )

    project.share([request], send_notification=send_notification)

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {
        "data": [{"username": "me@datarobot.com", "role": "EDITOR"}],
        "sendNotification": send_notification,
    }


@responses.activate
@pytest.mark.parametrize("include_feature_discovery_entities", [True, False])
def test_share_include_feature_discovery_entities(
    include_feature_discovery_entities, project_endpoint, project_server_data
):
    project = Project.from_server_data(project_server_data)
    request = SharingAccess("me@datarobot.com", "EDITOR")
    responses.add(
        responses.PATCH, "{}{}/accessControl/".format(project_endpoint, project.id), status=204
    )

    project.share([request], include_feature_discovery_entities=include_feature_discovery_entities)

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {
        "data": [{"username": "me@datarobot.com", "role": "EDITOR"}],
        "includeFeatureDiscoveryEntities": include_feature_discovery_entities,
    }


@responses.activate
def test_create_interaction_feature(project_endpoint, project, interaction_feature_server_data):
    responses.add(
        responses.POST,
        "{}{}/interactionFeatures/".format(project_endpoint, project.id),
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )

    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": "https://host_name.com/projects/{}/".format(project.id)},
    )

    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/".format(project.id),
        status=200,
        body=json.dumps(interaction_feature_server_data),
        content_type="application/json",
    )

    project.create_interaction_feature("int0", ["cat1", "cat2"], "+")

    assert responses.calls[0].response.status_code == 202
    assert responses.calls[1].response.status_code == 303
    assert responses.calls[2].response.status_code == 200

    assert request_body_to_json(responses.calls[0].request) == {
        "featureName": "int0",
        "features": ["cat1", "cat2"],
        "separator": "+",
    }


def test_create_interaction_feature_arguments_validation(project):
    with pytest.raises(TypeError):
        project.create_interaction_feature("name", "f1 & f2", "+")

    with pytest.raises(ValueError):
        project.create_interaction_feature("name", ["f1", "f2", "f2"], "+")


@responses.activate
def test_get_relationships_configuration(project, relationships_configuration):
    id = project.id
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/relationshipsConfiguration/".format(id),
        json=relationships_configuration,
    )

    result = project.get_relationships_configuration()
    assert responses.calls[0].request.method == "GET"
    assert result.id == relationships_configuration["id"]
    assert len(result.dataset_definitions) == 2
    assert len(result.relationships) == 2

    for actual, expected in zip(
        result.dataset_definitions, relationships_configuration["datasetDefinitions"]
    ):
        assert actual["identifier"] == expected["identifier"]
        assert actual["catalog_version_id"] == expected["catalogVersionId"]
        assert actual["catalog_id"] == expected["catalogId"]
        assert actual["snapshot_policy"] == expected["snapshotPolicy"]

    actual = result.relationships[1]
    expected = relationships_configuration["relationships"][1]
    assert actual["dataset1_identifier"] == expected["dataset1Identifier"]
    assert actual["dataset2_identifier"] == expected["dataset2Identifier"]
    assert actual["dataset1_keys"] == expected["dataset1Keys"]
    assert actual["dataset2_keys"] == expected["dataset2Keys"]
    assert actual["feature_derivation_window_end"] == expected["featureDerivationWindowEnd"]
    assert actual["feature_derivation_window_start"] == expected["featureDerivationWindowStart"]
    assert (
        actual["feature_derivation_window_time_unit"] == expected["featureDerivationWindowTimeUnit"]
    )
    assert actual["prediction_point_rounding"] == expected["predictionPointRounding"]
    assert (
        actual["prediction_point_rounding_time_unit"] == expected["predictionPointRoundingTimeUnit"]
    )


@responses.activate
@pytest.mark.parametrize("mode", [AUTOPILOT_MODE.FULL_AUTO])
@pytest.mark.parametrize("blend_best_models", [True, False])
@pytest.mark.parametrize("scoring_code_only", [True, False])
@pytest.mark.parametrize("prepare_model_for_deployment", [True, False])
def test_start_autopilot(
    project_endpoint,
    project,
    mode,
    blend_best_models,
    scoring_code_only,
    prepare_model_for_deployment,
):
    responses.add(
        responses.POST,
        "{}{}/autopilots/".format(project_endpoint, project.id),
        status=201,
        content_type="application/json",
    )

    project.start_autopilot(
        "featurelist-id",
        mode=mode,
        blend_best_models=blend_best_models,
        scoring_code_only=scoring_code_only,
        prepare_model_for_deployment=prepare_model_for_deployment,
    )
    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {
        "featurelistId": "featurelist-id",
        "mode": mode,
        "blendBestModels": blend_best_models,
        "scoringCodeOnly": scoring_code_only,
        "prepareModelForDeployment": prepare_model_for_deployment,
    }


@pytest.fixture
def dataset_content():
    return "some_random_file_contents" * 1000000


@pytest.fixture
def local_csv_filename(tmpdir):
    return str(tmpdir.join("temp.csv"))


@responses.activate
def test_feature_discovery_download_training_dataset(dataset_content, local_csv_filename):
    project_id = "pid"
    url = "https://host_name.com/projects/{}/featureDiscoveryDatasetDownload/".format(project_id)
    responses.add(responses.GET, url, body=dataset_content, status=200, match_querystring=True)
    project = Project(id=project_id)
    project.download_feature_discovery_dataset(local_csv_filename)
    assert os.path.isfile(local_csv_filename)
    with open(local_csv_filename) as f:
        contents = f.read()
        assert contents == dataset_content
    os.unlink(local_csv_filename)
    assert not os.path.isfile(local_csv_filename)


@responses.activate
def test_feature_discovery_download_prediction_dataset(dataset_content, local_csv_filename):
    project_id = "pid"
    dataset_id = "did"
    url = "https://host_name.com/projects/{}/featureDiscoveryDatasetDownload/?datasetId={}".format(
        project_id, dataset_id
    )
    responses.add(responses.GET, url, body=dataset_content, status=200, match_querystring=True)
    project = Project(id=project_id)
    project.download_feature_discovery_dataset(local_csv_filename, pred_dataset_id=dataset_id)
    assert os.path.isfile(local_csv_filename)
    with open(local_csv_filename) as f:
        contents = f.read()
        assert contents == dataset_content
    os.unlink(local_csv_filename)
    assert not os.path.isfile(local_csv_filename)


@pytest.fixture
def sql_content():
    return "some_random_sql_contents" * 1000000


@pytest.fixture
def local_sql_filename(tmpdir):
    return str(tmpdir.join("temp.sql"))


@responses.activate
def test_feature_discovery_download_project_sql(sql_content, local_sql_filename):
    pid = "pid"
    export_url = "https://host_name.com/projects/{}/featureDiscoveryRecipeSqlExports/".format(pid)
    download_url = "https://host_name.com/projects/{}/featureDiscoveryRecipeSQLs/download/"
    download_url = download_url.format(pid)

    responses.add(
        responses.POST,
        export_url,
        body="",
        status=202,
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )
    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": download_url},
    )
    responses.add(
        responses.GET, download_url, status=200, body=sql_content, content_type="application/json",
    )

    project = Project(id=pid)
    project.download_feature_discovery_recipe_sqls(local_sql_filename)
    assert os.path.isfile(local_sql_filename)
    with open(local_sql_filename) as f:
        contents = f.read()
        assert contents == sql_content
    os.unlink(local_sql_filename)
    assert not os.path.isfile(local_sql_filename)


@responses.activate
def test_feature_discovery_download_model_sql(sql_content, local_sql_filename):
    pid = "pid"
    model_id = "lid"
    export_url = "https://host_name.com/projects/{}/featureDiscoveryRecipeSqlExports/".format(pid)
    download_url = "https://host_name.com/projects/{}/featureDiscoveryRecipeSQLs/download/"
    download_url = download_url.format(pid)
    download_url = "{}?modelId={}".format(download_url, model_id)

    responses.add(
        responses.POST,
        export_url,
        body="",
        status=202,
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )
    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": download_url},
    )
    responses.add(
        responses.GET,
        download_url,
        status=200,
        body=sql_content,
        content_type="application/json",
        match_querystring=True,
    )

    project = Project(id=pid)
    project.download_feature_discovery_recipe_sqls(local_sql_filename, model_id=model_id)

    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == {"modelId": "lid"}

    assert os.path.isfile(local_sql_filename)
    with open(local_sql_filename) as f:
        contents = f.read()
        assert contents == sql_content
    os.unlink(local_sql_filename)
    assert not os.path.isfile(local_sql_filename)



import json

import pytest
import responses

from datarobot import Deployment
from datarobot.utils import from_api
from tests.utils import request_body_to_json


@pytest.fixture()
def deployment_settings_response_data():
    return {
        "targetDrift": {"enabled": True},
        "featureDrift": {"enabled": True},
        "associationId": {"columnNames": ["column"], "requiredInPredictionRequests": False},
        "predictionWarning": {"enabled": True, "customBoundaries": {"upper": 1337, "lower": 0}},
        "predictionIntervals": {"enabled": True, "percentiles": [80]},
        "predictionsDataCollection": {"enabled": True},
        "otherSettings": {"test": "value"},
    }


@pytest.fixture()
def deployment_settings_get_response(
    unittest_endpoint, deployment_data, deployment_settings_response_data
):
    deployment_id = deployment_data["id"]
    url = "{}/deployments/{}/settings/".format(unittest_endpoint, deployment_id)
    responses.add(
        responses.GET,
        url,
        body=json.dumps(deployment_settings_response_data),
        status=200,
        content_type="application/json",
    )


@pytest.fixture()
def deployment_settings_update_response(unittest_endpoint, deployment_data):
    deployment_id = deployment_data["id"]
    update_setting_url = "{}/deployments/{}/settings/".format(unittest_endpoint, deployment_id)
    deployment_url = "{}/deployments/{}/".format(unittest_endpoint, deployment_id)
    status_url = "{}/status_url".format(unittest_endpoint)
    responses.add(responses.PATCH, update_setting_url, headers={"Location": status_url}, status=202)
    responses.add(responses.GET, status_url, headers={"Location": deployment_url}, status=303)


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
def test_cannot_update_nothing(deployment_data):
    deployment_id = deployment_data["id"]
    deployment = Deployment.get(deployment_id)
    with pytest.raises(ValueError):
        deployment.update_drift_tracking_settings()


@responses.activate
@pytest.mark.usefixtures("deployment_get_response", "deployment_settings_get_response")
def test_get_drift_tracking_settings(
    unittest_endpoint, deployment_data, deployment_settings_response_data
):
    deployment_id = deployment_data["id"]
    deployment = Deployment.get(deployment_id)
    settings = deployment.get_drift_tracking_settings()

    response = from_api(deployment_settings_response_data)
    expected_settings = {
        "target_drift": response["target_drift"],
        "feature_drift": response["feature_drift"],
    }
    assert settings == expected_settings


@responses.activate
@pytest.mark.usefixtures("deployment_get_response", "deployment_settings_update_response")
@pytest.mark.parametrize(
    "param, expected_request",
    [
        ({"target_drift_enabled": True}, {"targetDrift": {"enabled": True}}),
        ({"feature_drift_enabled": False}, {"featureDrift": {"enabled": False}}),
        (
            {"target_drift_enabled": True, "feature_drift_enabled": True},
            {"targetDrift": {"enabled": True}, "featureDrift": {"enabled": True}},
        ),
        (
            {"target_drift_enabled": False, "feature_drift_enabled": False},
            {"targetDrift": {"enabled": False}, "featureDrift": {"enabled": False}},
        ),
    ],
)
def test_update_drift_tracking_settings(
    unittest_endpoint, deployment_data, param, expected_request
):
    deployment_id = deployment_data["id"]
    deployment = Deployment.get(deployment_id)
    deployment.update_drift_tracking_settings(**param)

    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body == expected_request


@responses.activate
@pytest.mark.usefixtures("deployment_get_response", "deployment_settings_get_response")
def test_get_association_id_settings(
    unittest_endpoint, deployment_data, deployment_settings_response_data
):
    deployment_id = deployment_data["id"]
    deployment = Deployment.get(deployment_id)
    settings = deployment.get_association_id_settings()

    response = from_api(deployment_settings_response_data)
    expected_settings = {
        "column_names": response["association_id"]["column_names"],
        "required_in_prediction_requests": response["association_id"][
            "required_in_prediction_requests"
        ],
    }
    assert settings == expected_settings


@responses.activate
@pytest.mark.usefixtures("deployment_get_response", "deployment_settings_update_response")
@pytest.mark.parametrize(
    "param, expected_request",
    [
        ({"column_names": ["a"]}, {"associationId": {"columnNames": ["a"]}}),
        (
            {"required_in_prediction_requests": True},
            {"associationId": {"requiredInPredictionRequests": True}},
        ),
        (
            {"column_names": ["a"], "required_in_prediction_requests": True},
            {"associationId": {"columnNames": ["a"], "requiredInPredictionRequests": True}},
        ),
    ],
)
def test_update_association_id_settings(
    unittest_endpoint, deployment_data, param, expected_request
):
    deployment_id = deployment_data["id"]
    deployment = Deployment.get(deployment_id)
    deployment.update_association_id_settings(**param)

    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body == expected_request


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
def test_get_prediction_warning_settings(
    unittest_endpoint, deployment_data, deployment_settings_response_data
):
    deployment_id = deployment_data["id"]
    url = "{}/deployments/{}/settings/".format(unittest_endpoint, deployment_id)
    responses.add(
        responses.GET,
        url,
        body=json.dumps(deployment_settings_response_data),
        status=200,
        content_type="application/json",
    )

    deployment = Deployment.get(deployment_id)
    settings = deployment.get_prediction_warning_settings()

    response = from_api(deployment_settings_response_data)
    assert settings == response["prediction_warning"]


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
@pytest.mark.parametrize(
    "param, expected_request",
    [
        (
            {
                "prediction_warning_enabled": True,
                "use_default_boundaries": False,
                "upper_boundary": 1337,
                "lower_boundary": 0,
            },
            {
                "predictionWarning": {
                    "enabled": True,
                    "customBoundaries": {"upper": 1337, "lower": 0},
                }
            },
        ),
        (
            {"prediction_warning_enabled": True, "use_default_boundaries": True},
            {"predictionWarning": {"enabled": True, "customBoundaries": None}},
        ),
        ({"prediction_warning_enabled": False}, {"predictionWarning": {"enabled": False}}),
    ],
)
def test_update_prediction_warning_settings(
    unittest_endpoint, deployment_data, param, expected_request
):
    deployment_id = deployment_data["id"]
    update_setting_url = "{}/deployments/{}/settings/".format(unittest_endpoint, deployment_id)
    deployment_url = "{}/deployments/{}/".format(unittest_endpoint, deployment_id)
    status_url = "{}/status_url".format(unittest_endpoint)
    responses.add(responses.PATCH, update_setting_url, headers={"Location": status_url}, status=202)
    responses.add(responses.GET, status_url, headers={"Location": deployment_url}, status=303)
    deployment = Deployment.get(deployment_id)
    deployment.update_prediction_warning_settings(**param)
    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body == expected_request


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
def test_get_predictions_data_collection_settings(
    unittest_endpoint, deployment_data, deployment_settings_response_data
):
    deployment_id = deployment_data["id"]
    url = "{}/deployments/{}/settings/".format(unittest_endpoint, deployment_id)
    responses.add(
        responses.GET,
        url,
        body=json.dumps(deployment_settings_response_data),
        status=200,
        content_type="application/json",
    )

    deployment = Deployment.get(deployment_id)
    settings = deployment.get_predictions_data_collection_settings()

    response = from_api(deployment_settings_response_data)
    assert settings == response["predictions_data_collection"]


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
@pytest.mark.parametrize(
    "param, expected_request",
    [
        ({"enabled": True}, {"predictionsDataCollection": {"enabled": True}}),
        ({"enabled": False}, {"predictionsDataCollection": {"enabled": False}}),
    ],
)
def test_update_predictions_data_collection_settings(
    unittest_endpoint, deployment_data, param, expected_request
):
    deployment_id = deployment_data["id"]
    update_setting_url = "{}/deployments/{}/settings/".format(unittest_endpoint, deployment_id)
    deployment_url = "{}/deployments/{}/".format(unittest_endpoint, deployment_id)
    status_url = "{}/status_url".format(unittest_endpoint)
    responses.add(
        responses.PATCH, update_setting_url, headers={"Location": status_url}, status=202,
    )
    responses.add(responses.GET, status_url, headers={"Location": deployment_url}, status=303)
    deployment = Deployment.get(deployment_id)
    deployment.update_predictions_data_collection_settings(**param)

    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body == expected_request


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
def test_get_prediction_intervals_settings(
    unittest_endpoint, deployment_data, deployment_settings_response_data
):
    deployment_id = deployment_data["id"]
    url = "{}/deployments/{}/settings/".format(unittest_endpoint, deployment_id)
    responses.add(
        responses.GET,
        url,
        body=json.dumps(deployment_settings_response_data),
        status=200,
        content_type="application/json",
    )

    deployment = Deployment.get(deployment_id)
    settings = deployment.get_prediction_intervals_settings()

    response = from_api(deployment_settings_response_data)
    expected_settings = response["prediction_intervals"]
    assert settings == expected_settings


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
def test_update_prediction_intervals_settings(unittest_endpoint, deployment_data):
    deployment_id = deployment_data["id"]
    update_setting_url = "{}/deployments/{}/settings/".format(unittest_endpoint, deployment_id)
    deployment_url = "{}/deployments/{}/".format(unittest_endpoint, deployment_id)
    status_url = "{}/status_url".format(unittest_endpoint)
    project_id = deployment_data["model"]["projectId"]
    model_id = deployment_data["model"]["id"]

    # responses for ensuring percentile is calculated for model
    responses.add(
        responses.POST,
        "{}/projects/{}/models/{}/predictionIntervals/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=202,
        headers={"Location": "{}/projects/{}/jobs/1/".format(unittest_endpoint, project_id)},
    )
    responses.add(
        responses.GET,
        "{}/projects/{}/jobs/1/".format(unittest_endpoint, project_id),
        status=303,
        content_type="application/json",
        body=json.dumps(
            {
                "status": u"COMPLETED",
                "model_id": model_id,
                "is_blocked": False,
                "url": "{}/projects/{}/jobs/1/".format(unittest_endpoint, project_id),
                "job_type": "calculate_prediction_intervals",
                "project_id": project_id,
                "id": 1,
            }
        ),
        headers={"Location": "{}/projects/{}/jobs/1/result/".format(unittest_endpoint, project_id)},
    )

    # responses for updating deployment settings
    responses.add(responses.PATCH, update_setting_url, headers={"Location": status_url}, status=202)
    responses.add(responses.GET, status_url, headers={"Location": deployment_url}, status=303)

    deployment = Deployment.get(deployment_id)
    deployment.update_prediction_intervals_settings(percentiles=[70], enabled=False)

    request_body = request_body_to_json(responses.calls[5].request)
    assert request_body == {"predictionIntervals": {"enabled": False, "percentiles": [70]}}



import json

import responses

from datarobot import enums, ShapMatrixJob
from datarobot.models.shap_matrix import ShapMatrix
from tests.test_job import mock_job_data


@responses.activate
def test_shap_matrix_job__refresh__ok():
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.ERROR)
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=200,
        body=json.dumps(refresh_job_data),
        content_type="application/json",
    )

    job = ShapMatrixJob(mock_job_data, model_id="model-id", dataset_id="dataset-id")
    job.refresh()

    assert job.status == enums.QUEUE_STATUS.ERROR


@responses.activate
def test_shap_matrix_job__wait_for_completion__ok():
    resource_url = "https://host_name.com/projects/p-id/shapMatrices/shap-matrix-id"
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.COMPLETED)
    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=303,
        body=json.dumps(refresh_job_data),
        content_type="application/json",
        adding_headers={"Location": resource_url},
    )

    job = ShapMatrixJob(mock_job_data, model_id="model-id", dataset_id="dataset-id")
    job.wait_for_completion(max_wait=3)

    assert job.status == enums.QUEUE_STATUS.COMPLETED


@responses.activate
def test_get_completed_shap_matrix_with_shap_matrix_job(client):
    job_id = "1"
    project_id = "p-id"
    shap_matrix_id = "shap-matrix-id"
    data_path = "projects/{project_id}/shapMatrices/{shap_matrix_id}/".format(
        project_id=project_id, shap_matrix_id=shap_matrix_id
    )
    get_shap_matrix_url = "https://host_name.com/{data_path}".format(data_path=data_path)
    mock_generic_job_data = {
        "status": enums.QUEUE_STATUS.COMPLETED,
        "url": "https://host_name.com/projects/p-id/predictJobs/1/",
        "id": job_id,
        "jobType": "a job",
        "isBlocked": False,
        "projectId": project_id,
    }

    responses.add(
        responses.GET,
        "https://host_name.com/projects/p-id/jobs/1/",
        status=303,
        body=json.dumps(dict(mock_generic_job_data)),
        content_type="application/json",
        adding_headers={"Location": get_shap_matrix_url},
    )

    job = ShapMatrixJob.get(project_id, job_id)

    data = job.get_result()
    assert job.status == "COMPLETED"

    assert isinstance(data, ShapMatrix)
    assert data.project_id == project_id
    assert data.id == shap_matrix_id
    assert data.model_id is None
    assert data.dataset_id is None



import trafaret as t

from datarobot.enums import FEATURE_ASSOCIATION_METRIC, FEATURE_ASSOCIATION_TYPE
from datarobot.models.api_object import APIObject
from datarobot.utils import encode_utf8_if_py2


class FeatureAssociationMatrix(APIObject):
    """
    Feature association statistics for a project.

    .. note::
        Projects created prior to v2.17 are not supported by this feature.

    Attributes
    ----------
    project_id : str
        Id of the associated project.
    strengths : list of dict
        Pairwise statistics for the available features as structured below.
    features : list of dict
        Metadata for each feature and where it goes in the matrix.

    Examples
    --------
    .. code-block:: python

        import datarobot as dr

        # retrieve feature association matrix
        feature_association_matrix = dr.FeatureAssociationMatrix.get(project_id)
        feature_association_matrix.strength
        feature_association_matrix.features

        # retrieve feature association matrix for a metric, association type or a feature list
        feature_association_matrix = dr.FeatureAssociationMatrix.get(
            project_id,
            metric=enums.FEATURE_ASSOCIATION_METRIC.SPEARMAN,
            association_type=enums.FEATURE_ASSOCIATION_TYPE.CORRELATION,
            featurelist_id=featurelist_id,
        )
    """

    _path = "projects/{}/featureAssociationMatrix/"
    _association_strength = t.Dict(
        {
            t.Key("feature1"): t.String(allow_blank=True),
            t.Key("feature2"): t.String(allow_blank=True),
            t.Key("statistic"): t.Float(),
        }
    )
    _association_feature = t.Dict(
        {
            t.Key("cluster_sort_index", optional=True): t.Int(),
            t.Key("cluster_name", optional=True): t.String(),
            t.Key("cluster_id", optional=True): t.Int(),
            t.Key("feature"): t.String(allow_blank=True),
            t.Key("strength_sort_index"): t.Int(),
            t.Key("alphabetic_sort_index"): t.Int(),
            t.Key("importance_sort_index"): t.Int(),
        }
    )
    _converter = t.Dict(
        {
            t.Key("strengths"): t.List(_association_strength),
            t.Key("features"): t.List(_association_feature),
        }
    )
    _query_param_validator = t.Dict(
        {
            t.Key("metric", optional=True): t.Enum(*FEATURE_ASSOCIATION_METRIC.ALL),
            t.Key("type", optional=True): t.Enum(*FEATURE_ASSOCIATION_TYPE.ALL),
            t.Key("featurelistId", optional=True): t.String() | t.Null(),
        }
    )

    def __init__(self, strengths=None, features=None, project_id=None):
        self.strengths = strengths
        self.features = features
        self.project_id = project_id

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}(project_id={}, strengths={}, features={})".format(
                self.__class__.__name__, self.project_id, self.strengths, self.features,
            )
        )

    @classmethod
    def get(cls, project_id, metric=None, association_type=None, featurelist_id=None):
        """
        Get feature association statistics.

        Parameters
        ----------
        project_id : str
            Id of the project that contains the requested associations.
        metric : enums.FEATURE_ASSOCIATION_METRIC
            The name of a metric to get pairwise data for. Since 'v2.19' this is optional and
            defaults to `enums.FEATURE_ASSOCIATION_METRIC.MUTUAL_INFO`.
        association_type : enums.FEATURE_ASSOCIATION_TYPE
            The type of dependence for the data. Since 'v2.19' this is optional and defaults to
            `enums.FEATURE_ASSOCIATION_TYPE.ASSOCIATION`.
        featurelist_id : str or None
            Optional, the feature list to lookup FAM data for. By default, depending on the type of
            the project "Informative Features" or "Timeseries Informative Features" list will be
            used.
            (New in version v2.19)

        Returns
        -------
        FeatureAssociationMatrix
            Feature association pairwise metric strength data, clustering data, and ordering data
            for Feature Association Matrix visualization.
        """
        t.String().check(project_id)
        params = dict()
        if metric:
            params["metric"] = metric
        if association_type:
            params["type"] = association_type
        if featurelist_id:
            params["featurelistId"] = featurelist_id
        cls._query_param_validator.check(params)

        url = cls._path.format(project_id)
        response = cls._client.get(url, params=params)
        feature_association_matrix = cls.from_server_data(response.json())
        # FAM public API doesn't include project_id so lets populate it
        feature_association_matrix.project_id = project_id
        return feature_association_matrix

    def to_dict(self):
        return {"strengths": self.strengths, "features": self.features}



from .feature_discovery import *  # noqa
from .partitioning_methods import *  # noqa


class AdvancedOptions(object):
    """
    Used when setting the target of a project to set advanced options of modeling process.

    Parameters
    ----------
    weights : string, optional
        The name of a column indicating the weight of each row
    response_cap : float in [0.5, 1), optional
        Quantile of the response distribution to use for response capping.
    blueprint_threshold : int, optional
        Number of hours models are permitted to run before being excluded from later autopilot
        stages
        Minimum 1
    seed : int
        a seed to use for randomization
    smart_downsampled : bool
        whether to use smart downsampling to throw away excess rows of the majority class.  Only
        applicable to classification and zero-boosted regression projects.
    majority_downsampling_rate : float
        the percentage between 0 and 100 of the majority rows that should be kept.  Specify only if
        using smart downsampling.  May not cause the majority class to become smaller than the
        minority class.
    offset : list of str, optional
        (New in version v2.6) the list of the names of the columns containing the offset
        of each row
    exposure : string, optional
        (New in version v2.6) the name of a column containing the exposure of each row
    accuracy_optimized_mb : bool, optional
        (New in version v2.6) Include additional, longer-running models that will be run by the
        autopilot and available to run manually.
    scaleout_modeling_mode : string, optional
        (New in version v2.8) Specifies the behavior of Scaleout models for the project.
        This is one of ``datarobot.enums.SCALEOUT_MODELING_MODE``.
        If ``datarobot.enums.SCALEOUT_MODELING_MODE.DISABLED``, no
        models will run during autopilot or show in the list of available blueprints.
        Scaleout models must be disabled for some partitioning settings including projects
        using datetime partitioning or projects using offset or exposure columns.
        If ``datarobot.enums.SCALEOUT_MODELING_MODE.REPOSITORY_ONLY``,
        scaleout models will be in the list of available blueprints
        but not run during autopilot.
        If ``datarobot.enums.SCALEOUT_MODELING_MODE.AUTOPILOT``,
        scaleout models will run during autopilot and be in the list of
        available blueprints.
        Scaleout models are only supported in the Hadoop enviroment with
        the corresponding user permission set.
    events_count : string, optional
        (New in version v2.8) the name of a column specifying events count.
    monotonic_increasing_featurelist_id : string, optional
        (new in version 2.11) the id of the featurelist that defines the set of features
        with a monotonically increasing relationship to the target. If None,
        no such constraints are enforced. When specified, this will set a default for the project
        that can be overriden at model submission time if desired.
    monotonic_decreasing_featurelist_id : string, optional
        (new in version 2.11) the id of the featurelist that defines the set of features
        with a monotonically decreasing relationship to the target. If None,
        no such constraints are enforced. When specified, this will set a default for the project
        that can be overriden at model submission time if desired.
    only_include_monotonic_blueprints : bool, optional
        (new in version 2.11) when true, only blueprints that support enforcing
        monotonic constraints will be available in the project or selected for the autopilot.
    allowed_pairwise_interaction_groups : list of tuple, optional
        (New in version v2.19) For GAM models - specify groups of columns for which pairwise
        interactions will be allowed. E.g. if set to [(A, B, C), (C, D)] then GAM models will
        allow interactions between columns AxB, BxC, AxC, CxD. All others (AxD, BxD) will
        not be considered.
    blend_best_models: bool, optional
        (New in version v2.19) blend best models during Autopilot run
    scoring_code_only: bool, optional
        (New in version v2.19) Keep only models that can be converted to scorable java code
        during Autopilot run
    shap_only_mode: bool, optional
        (New in version v2.21) Keep only models that support SHAP values during Autopilot run. Use
        SHAP-based insights wherever possible. Defaults to False.
    prepare_model_for_deployment: bool, optional
        (New in version v2.19) Prepare model for deployment during Autopilot run.
        The preparation includes creating reduced feature list models, retraining best model
        on higher sample size, computing insights and assigning "RECOMMENDED FOR DEPLOYMENT" label.
    consider_blenders_in_recommendation: bool, optional
        (New in version 2.22.0) Include blenders when selecting a model to prepare for
        deployment in an Autopilot Run. Defaults to False.
    min_secondary_validation_model_count: int, optional
        (New in version v2.19) Compute "All backtest" scores (datetime models) or cross validation
        scores for the specified number of highest ranking models on the Leaderboard,
        if over the Autopilot default.
    autopilot_data_sampling_method: str, optional
        (New in version v2.23) one of ``datarobot.enums.DATETIME_AUTOPILOT_DATA_SAMPLING_METHOD``.
        Applicable for OTV projects only, defines if autopilot uses "random" or "latest" sampling
        when iteratively building models on various training samples. Defaults to "random" for
        duration-based projects and to "latest" for row-based projects.
    run_leakage_removed_feature_list: bool, optional
        (New in version v2.23) Run Autopilot on Leakage Removed feature list (if exists).
    autopilot_with_feature_discovery: bool, optional.
        (New in version v2.23) If true, autopilot will run on a feature list that includes features
        found via search for interactions.
    feature_discovery_supervised_feature_reduction: bool,  default ``True` optional
        (New in version v2.23) Run supervised feature reduction for feature discovery projects.

    Examples
    --------
    .. code-block:: python

        import datarobot as dr
        advanced_options = dr.AdvancedOptions(
            weights='weights_column',
            offset=['offset_column'],
            exposure='exposure_column',
            response_cap=0.7,
            blueprint_threshold=2,
            smart_downsampled=True, majority_downsampling_rate=75.0)

    """

    def __init__(
        self,
        weights=None,
        response_cap=None,
        blueprint_threshold=None,
        seed=None,
        smart_downsampled=False,
        majority_downsampling_rate=None,
        offset=None,
        exposure=None,
        accuracy_optimized_mb=None,
        scaleout_modeling_mode=None,
        events_count=None,
        monotonic_increasing_featurelist_id=None,
        monotonic_decreasing_featurelist_id=None,
        only_include_monotonic_blueprints=None,
        allowed_pairwise_interaction_groups=None,
        blend_best_models=None,
        scoring_code_only=None,
        prepare_model_for_deployment=None,
        consider_blenders_in_recommendation=None,
        min_secondary_validation_model_count=None,
        shap_only_mode=None,
        autopilot_data_sampling_method=None,
        run_leakage_removed_feature_list=None,
        autopilot_with_feature_discovery=False,
        feature_discovery_supervised_feature_reduction=True,
    ):
        self.weights = weights
        self.response_cap = response_cap
        self.blueprint_threshold = blueprint_threshold
        self.seed = seed
        self.smart_downsampled = smart_downsampled
        self.majority_downsampling_rate = majority_downsampling_rate
        self.offset = offset
        self.exposure = exposure
        self.accuracy_optimized_mb = accuracy_optimized_mb
        self.scaleout_modeling_mode = scaleout_modeling_mode
        self.events_count = events_count
        self.monotonic_increasing_featurelist_id = monotonic_increasing_featurelist_id
        self.monotonic_decreasing_featurelist_id = monotonic_decreasing_featurelist_id
        self.only_include_monotonic_blueprints = only_include_monotonic_blueprints
        self.allowed_pairwise_interaction_groups = allowed_pairwise_interaction_groups
        self.blend_best_models = blend_best_models
        self.scoring_code_only = scoring_code_only
        self.shap_only_mode = shap_only_mode
        self.prepare_model_for_deployment = prepare_model_for_deployment
        self.consider_blenders_in_recommendation = consider_blenders_in_recommendation
        self.min_secondary_validation_model_count = min_secondary_validation_model_count
        self.autopilot_data_sampling_method = autopilot_data_sampling_method
        self.run_leakage_removed_feature_list = run_leakage_removed_feature_list
        self.autopilot_with_feature_discovery = autopilot_with_feature_discovery
        self.feature_discovery_supervised_feature_reduction = (
            feature_discovery_supervised_feature_reduction
        )

    def collect_payload(self):

        payload = dict(
            weights=self.weights,
            response_cap=self.response_cap,
            blueprint_threshold=self.blueprint_threshold,
            seed=self.seed,
            smart_downsampled=self.smart_downsampled,
            majority_downsampling_rate=self.majority_downsampling_rate,
            offset=self.offset,
            exposure=self.exposure,
            accuracy_optimized_mb=self.accuracy_optimized_mb,
            scaleout_modeling_mode=self.scaleout_modeling_mode,
            events_count=self.events_count,
            monotonic_increasing_featurelist_id=self.monotonic_increasing_featurelist_id,
            monotonic_decreasing_featurelist_id=self.monotonic_decreasing_featurelist_id,
            only_include_monotonic_blueprints=self.only_include_monotonic_blueprints,
            allowed_pairwise_interaction_groups=self.allowed_pairwise_interaction_groups,
        )

        # Some of the optional parameters are incompatible with the others.
        # For example, scoring_code_only is not compatible with scaleout_modeling_mode.
        # Api will return 422 if both parameters are present.
        sfd = self.feature_discovery_supervised_feature_reduction
        optional = dict(
            blend_best_models=self.blend_best_models,
            scoring_code_only=self.scoring_code_only,
            shap_only_mode=self.shap_only_mode,
            prepare_model_for_deployment=self.prepare_model_for_deployment,
            consider_blenders_in_recommendation=self.consider_blenders_in_recommendation,
            min_secondary_validation_model_count=self.min_secondary_validation_model_count,
            autopilot_data_sampling_method=self.autopilot_data_sampling_method,
            run_leakage_removed_feature_list=self.run_leakage_removed_feature_list,
            autopilot_with_feature_discovery=self.autopilot_with_feature_discovery,
            feature_discovery_supervised_feature_reduction=sfd,
        )

        payload.update({k: v for k, v in optional.items() if v is not None})

        return payload



from operator import itemgetter

import trafaret as t

from datarobot.models.api_object import APIObject

from ..enums import AnomalyAssessmentStatus, SOURCE_TYPE
from ..utils import encode_utf8_if_py2, from_api
from ..utils.pagination import unpaginate
from ..utils.waiters import wait_for_async_resolution

DEFAULT_BATCH_SIZE = 1000

RecordMetadataTrafaret = t.Dict(
    {
        t.Key("record_id"): t.String,
        t.Key("project_id"): t.String,
        t.Key("model_id"): t.String,
        t.Key("backtest"): t.String() | t.Int,
        t.Key("source"): t.Enum(*SOURCE_TYPE.ALL),
        t.Key("series_id"): t.String() | t.Null,
    }
)


class BaseAPIObject(APIObject):
    def __init__(self, **record_kwargs):
        self.record_id = record_kwargs["record_id"]
        self.project_id = record_kwargs["project_id"]
        self.model_id = record_kwargs["model_id"]
        self.backtest = record_kwargs["backtest"]
        self.source = record_kwargs["source"]
        self.series_id = record_kwargs["series_id"]

    def __repr__(self):
        return encode_utf8_if_py2(
            "{}(project_id={}, model_id={}, series_id={}, "
            "backtest={}, source={}, record_id={})".format(
                self.__class__.__name__,
                self.project_id,
                self.model_id,
                self.series_id,
                self.backtest,
                self.source,
                self.record_id,
            )
        )

    @classmethod
    def from_server_data(cls, data, keep_attrs=None):
        """
        Instantiate an object of this class using the data directly from the server,
        meaning that the keys may have the wrong camel casing

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        keep_attrs : list
            List of the dotted namespace notations for attributes to keep within the
            object structure even if their values are None
        """
        case_converted = from_api(data, keep_attrs=keep_attrs, keep_null_keys=True)
        return cls.from_data(case_converted)


class AnomalyAssessmentRecord(BaseAPIObject):
    """Object which keeps metadata about anomaly assessment insight for the particular
    subset, backtest and series and the links to proceed to get the anomaly assessment data.

    .. versionadded:: v2.25

    Attributes
    ----------
    record_id: str
        The ID of the record.
    project_id: str
        The ID of the project record belongs to.
    model_id: str
        The ID of the model record belongs to.
    backtest: int or "holdout"
        The backtest of the record.
    source: "training" or "validation"
        The source of the record
    series_id: str or None
        The series id of the record for the multiseries projects. Defined only for the multiseries
        projects.
    status: str
        The status of the insight. One of ``datarobot.enums.AnomalyAssessmentStatus``
    status_details: str
        The explanation of the status.
    start_date: str or None
        See start_date info in `Notes` for more details.
    end_date: str or None
        See end_date info in `Notes` for more details.
    prediction_threshold: float or None
        See prediction_threshold info in `Notes` for more details.
    preview_location: str or None
        See preview_location info in `Notes` for more details.
    latest_explanations_location: str or None
        See latest_explanations_location info in `Notes` for more details.
    delete_location: str
        The URL to delete anomaly assessment record and relevant insight data.

    Notes
    -----

    ``Record`` contains:

    * ``record_id`` : the ID of the record.
    * ``project_id`` : the project ID of the record.
    * ``model_id`` : the model ID of the record.
    * ``backtest`` : the backtest of the record.
    * ``source`` : the source of the record.
    * ``series_id`` : the series id of the record for the multiseries projects.
    * ``status`` : the status of the insight.
    * ``status_details`` : the explanation of the status.
    * ``start_date`` : the ISO-formatted timestamp of the first prediction in the subset. Will be
      None if status is not `AnomalyAssessmentStatus.COMPLETED`.
    * ``end_date`` : the ISO-formatted timestamp of the last prediction in the subset. Will be None
      if status is not `AnomalyAssessmentStatus.COMPLETED`.
    * ``prediction_threshold`` : the threshold, all rows with anomaly scores greater or equal to it
      have shap explanations computed.
      Will be None if status is not `AnomalyAssessmentStatus.COMPLETED`.
    * ``preview_location`` :  URL to retrieve predictions preview for the subset. Will be None if
      status is not `AnomalyAssessmentStatus.COMPLETED`.
    * ``latest_explanations_location`` : the URL to retrieve the latest predictions with
      the shap explanations. Will be None if status is not `AnomalyAssessmentStatus.COMPLETED`.
    * ``delete_location`` : the URL to delete anomaly assessment record and relevant insight data.

    """

    _path = "projects/{project_id}/anomalyAssessmentRecords/"
    _create_path = "projects/{project_id}/models/{model_id}/anomalyAssessmentInitialization/"

    _converter = (
        t.Dict(
            {
                t.Key("status"): t.Enum(*AnomalyAssessmentStatus.ALL),
                t.Key("status_details"): t.String,
                t.Key("start_date"): t.String() | t.Null,
                t.Key("end_date"): t.String() | t.Null,
                t.Key("prediction_threshold"): t.Float | t.Null,
                t.Key("preview_location"): t.String() | t.Null,
                t.Key("delete_location"): t.String(),
                t.Key("latest_explanations_location"): t.String() | t.Null,
            }
        )
        .merge(RecordMetadataTrafaret)
        .ignore_extra("*")
    )

    def __init__(
        self,
        status,
        status_details,
        start_date,
        end_date,
        prediction_threshold,
        preview_location,
        delete_location,
        latest_explanations_location,
        **record_kwargs
    ):
        self.status = status
        self.status_details = status_details
        self.start_date = start_date
        self.end_date = end_date
        self.prediction_threshold = prediction_threshold
        self.preview_location = preview_location
        self.delete_location = delete_location
        self.latest_explanations_location = latest_explanations_location
        super(AnomalyAssessmentRecord, self).__init__(**record_kwargs)

    @classmethod
    def list(
        cls,
        project_id,
        model_id,
        backtest=None,
        source=None,
        series_id=None,
        limit=100,
        offset=0,
        with_data_only=False,
    ):
        """Retrieve the list of the anomaly assessment records for the project and model.
        Output can be filtered and limited.

        Parameters
        ----------
        project_id: str
            The ID of the project record belongs to.
        model_id: str
            The ID of the model record belongs to.
        backtest: int or "holdout"
            The backtest to filter records by.
        source: "training" or "validation"
            The source to filter records by.
        series_id: str, optional
            The series id to filter records by. Can be specified for multiseries projects.
        limit: int, optional
            100 by default. At most this many results are returned.
        offset: int, optional
            This many results will be skipped.
        with_data_only: bool, False by default
            Filter by `status` == AnomalyAssessmentStatus.COMPLETED. If True, records with
            no data or not supported will be omitted.

        Returns
        -------
        AnomalyAssessmentRecord
            The anomaly assessment record.
        """
        params = {"limit": limit, "offset": offset}
        if model_id:
            params["modelId"] = model_id
        if backtest:
            params["backtest"] = backtest
        if source:
            params["source"] = source
        if series_id:
            params["series_id"] = series_id
        url = cls._path.format(project_id=project_id)
        if limit == 0:  # unlimited results
            params["limit"] = DEFAULT_BATCH_SIZE
            records = [
                cls.from_server_data(entry) for entry in unpaginate(url, params, cls._client)
            ]
        else:
            r_data = cls._client.get(url, params=params).json()
            records = [cls.from_server_data(item) for item in r_data["data"]]
        if with_data_only:
            records = [
                record for record in records if record.status == AnomalyAssessmentStatus.COMPLETED
            ]
        return records

    @classmethod
    def compute(cls, project_id, model_id, backtest, source, series_id=None):
        """Request anomaly assessment insight computation on the specified subset.

        Parameters
        ----------
        project_id: str
            The ID of the project to compute insight for.
        model_id: str
            The ID of the model to compute insight for.
        backtest: int or "holdout"
            The backtest to compute insight for.
        source: "training" or "validation"
            The source  to compute insight for.
        series_id: str, optional
            The series id to compute insight for. Required for multiseries projects.

        Returns
        -------
        AnomalyAssessmentRecord
            The anomaly assessment record.
        """
        payload = {"backtest": backtest, "source": source}
        if series_id:
            payload["series_id"] = series_id
        url = cls._create_path.format(project_id=project_id, model_id=model_id)
        response = cls._client.post(url, data=payload)
        finished_url = wait_for_async_resolution(cls._client, response.headers["Location"])
        r_data = cls._client.get(finished_url).json()
        # it ll be always one record
        return cls.from_server_data(r_data["data"][0])

    def delete(self):
        """ Delete anomaly assessment record with preview and explanations. """
        self._client.delete(self.delete_location)

    def get_predictions_preview(self):
        """Retrieve aggregated predictions statistics for the anomaly assessment record.

        Returns
        -------
        AnomalyAssessmentPredictionsPreview
        """
        data = self._client.get(self.preview_location).json()
        return AnomalyAssessmentPredictionsPreview.from_server_data(data)

    def get_latest_explanations(self):
        """Retrieve latest predictions along with shap explanations for the most anomalous records.

        Returns
        -------
        AnomalyAssessmentExplanations
        """
        data = self._client.get(self.latest_explanations_location).json()
        return AnomalyAssessmentExplanations.from_server_data(data)

    def get_explanations(self, start_date=None, end_date=None, points_count=None):
        """Retrieve predictions along with shap explanations for the most anomalous records
        in the specified date range/for defined number of points.
        Two out of three parameters: start_date, end_date or points_count must be specified.

        Parameters
        ----------
        start_date: str, optional
            The start of the date range to get explanations in.
            Example: ``2020-01-01T00:00:00.000000Z``
        end_date: str, optional
            The end of the date range to get explanations in.
            Example: ``2020-10-01T00:00:00.000000Z``
        points_count: int, optional
            The number of the rows to return.

        Returns
        -------
        AnomalyAssessmentExplanations
        """
        return AnomalyAssessmentExplanations.get(
            self.project_id,
            self.record_id,
            start_date=start_date,
            end_date=end_date,
            points_count=points_count,
        )

    def get_explanations_data_in_regions(self, regions, prediction_threshold=0.0):
        """Get predictions along with explanations for the specified regions, sorted by
        predictions in descending order.

        Parameters
        ----------
        regions: list of preview_bins
            For each region explanations will be retrieved and merged.
        prediction_threshold: float, optional
            If specified, only points with score greater or equal to the threshold will be returned.

        Returns
        -------
        dict in a form of {'explanations': explanations, 'shap_base_value': shap_base_value}

        """
        explanations = []
        shap_base_value = None
        for region in regions:
            response = self.get_explanations(
                start_date=region["start_date"], end_date=region["end_date"]
            )
            shap_base_value = response.shap_base_value
            for item in response.data:
                if item["prediction"] >= prediction_threshold:
                    explanations.append(item)
        explanations = list(sorted(explanations, key=itemgetter("prediction"), reverse=True))
        return {"explanations": explanations, "shap_base_value": shap_base_value}


class AnomalyAssessmentPredictionsPreview(BaseAPIObject):
    """Aggregated predictions over time for the corresponding anomaly assessment record.
    Intended to find the bins with highest anomaly scores.

    .. versionadded:: v2.25


    Attributes
    ----------
    record_id: str
        The ID of the record.
    project_id: str
        The ID of the project record belongs to.
    model_id: str
        The ID of the model record belongs to.
    backtest: int or "holdout"
        The backtest of the record.
    source: "training" or "validation"
        The source of the record
    series_id: str or None
        The series id of the record for the multiseries projects. Defined only for the multiseries
        projects.
    start_date: str
        the ISO-formatted timestamp of the first prediction in the subset.
    end_date: str
        the ISO-formatted timestamp of the last prediction in the subset.
    preview_bins:  list of preview_bin objects.
        The aggregated predictions for the subset. See more info in `Notes`.

    Notes
    -----

    ``AnomalyAssessmentPredictionsPreview`` contains:

    * ``record_id`` : the id of the corresponding anomaly assessment record.
    * ``project_id`` : the project ID of the corresponding anomaly assessment record.
    * ``model_id`` : the model ID of the corresponding anomaly assessment record.
    * ``backtest`` : the backtest of the corresponding anomaly assessment record.
    * ``source`` : the source of the corresponding anomaly assessment record.
    * ``series_id`` : the series id of the corresponding anomaly assessment record
      for the multiseries projects.
    * ``start_date`` : the  ISO-formatted timestamp of the first prediction in the subset.
    * ``end_date`` : the ISO-formatted timestamp of the last prediction in the subset.
    * ``preview_bins`` :  list of PreviewBin objects. The aggregated predictions for the subset.
      Bins boundaries may differ from actual start/end dates because this is an aggregation.

    ``PreviewBin`` contains:


    * ``start_date`` (str) : the ISO-formatted datetime of the start of the bin.
    * ``end_date`` (str) : the ISO-formatted datetime of the end of the bin.
    * ``avg_predicted`` (float or None) : the average prediction of the model in the bin. None if
      there are no entries in the bin.
    * ``max_predicted`` (float or None) : the maximum prediction of the model in the bin. None if
      there are no entries in the bin.
    * ``frequency`` (int) : the number of the rows in the bin.
    """

    _path = "projects/{project_id}/anomalyAssessmentRecords/{record_id}/predictionsPreview/"

    PreviewBinTrafaret = t.Dict(
        {
            t.Key("avg_predicted"): t.Float() | t.Null,
            t.Key("max_predicted"): t.Float() | t.Null,
            t.Key("start_date"): t.String,
            t.Key("end_date"): t.String,
            t.Key("frequency"): t.Int,
        }
    )

    _converter = (
        t.Dict(
            {
                t.Key("start_date"): t.String,
                t.Key("end_date"): t.String,
                t.Key("preview_bins"): t.List(PreviewBinTrafaret),
            }
        )
        .merge(RecordMetadataTrafaret)
        .ignore_extra("*")
    )

    def __init__(self, start_date, end_date, preview_bins, **record_kwargs):
        self.preview_bins = preview_bins
        self.start_date = start_date
        self.end_date = end_date
        super(AnomalyAssessmentPredictionsPreview, self).__init__(**record_kwargs)

    @classmethod
    def get(cls, project_id, record_id):
        """Retrieve aggregated predictions over time.

        Parameters
        ----------
        project_id: str
            The ID of the project.
        record_id: str
            The ID of the anomaly assessment record.

        Returns
        -------
        AnomalyAssessmentPredictionsPreview

        """
        url = cls._path.format(project_id=project_id, record_id=record_id)
        r_data = cls._client.get(url).json()
        return cls.from_server_data(r_data)

    def find_anomalous_regions(self, max_prediction_threshold=0.0):
        """Sort preview bins by max_predicted value and select those with max predicted value
         greater or equal to max prediction threshold.
         Sort the result by max predicted value in descending order.

        Parameters
        ----------
        max_prediction_threshold: float, optional
            Return bins with maximum anomaly score greater or equal to max_prediction_threshold.

        Returns
        -------
        preview_bins: list of preview_bin
            Filtered and sorted preview bins

        """
        no_empty_bins = [bin for bin in self.preview_bins if bin["frequency"]]
        filtered_bins = [
            bin for bin in no_empty_bins if bin["max_predicted"] >= max_prediction_threshold
        ]
        sorted_bins = list(sorted(filtered_bins, key=itemgetter("max_predicted"), reverse=True))
        return sorted_bins


class AnomalyAssessmentExplanations(BaseAPIObject):
    """Object which keeps predictions along with shap explanations for the most anomalous records
    in the specified date range/for defined number of points.

    .. versionadded:: v2.25

    Attributes
    ----------
    record_id: str
        The ID of the record.
    project_id: str
        The ID of the project record belongs to.
    model_id: str
        The ID of the model record belongs to.
    backtest: int or "holdout"
        The backtest of the record.
    source: "training" or "validation"
        The source of the record.
    series_id: str or None
        The series id of the record for the multiseries projects. Defined only for the multiseries
        projects.
    start_date: str or None
        The ISO-formatted datetime of the first row in the ``data``.
    end_date: str or None
        The ISO-formatted datetime of the last row in the ``data``.
    data: array of `data_point` objects or None
        See `data` info in `Notes` for more details.
    shap_base_value: float
        Shap base value.
    count: int
        The number of points in the ``data``.

    Notes
    -----

    ``AnomalyAssessmentExplanations`` contains:

    * ``record_id`` : the id of the corresponding anomaly assessment record.
    * ``project_id`` : the project ID of the corresponding anomaly assessment record.
    * ``model_id`` : the model ID of the corresponding anomaly assessment record.
    * ``backtest`` : the backtest of the corresponding anomaly assessment record.
    * ``source`` : the source of the corresponding anomaly assessment record.
    * ``series_id`` : the series id of the corresponding anomaly assessment record
      for the multiseries projects.
    * ``start_date`` : the ISO-formatted first timestamp in the response.
      Will be None of there is no data in the specified range.
    * ``end_date`` : the ISO-formatted last timestamp in the response.
      Will be None of there is no data in the specified range.
    * ``count`` : The number of points in the response.
    * ``shap_base_value`` : the shap base value.
    * ``data`` :  list of DataPoint objects in the specified date range.

    ``DataPoint`` contains:

     * ``shap_explanation`` : None or an array of up to 10 ShapleyFeatureContribution objects.
       Only rows with the highest anomaly scores have Shapley explanations calculated.
       Value is None if prediction is lower than `prediction_threshold`.
     * ``timestamp`` (str) : ISO-formatted timestamp for the row.
     * ``prediction`` (float) : The output of the model for this row.

    ``ShapleyFeatureContribution`` contains:

     * ``feature_value`` (str) : the feature value for this row. First 50 characters are returned.
     * ``strength`` (float) : the shap value for this feature and row.
     * ``feature`` (str) : the feature name.

    """

    _path = "projects/{project_id}/anomalyAssessmentRecords/{record_id}/explanations/"

    ShapContributionTrafaret = t.Dict(
        {t.Key("feature_value"): t.String, t.Key("strength"): t.Float, t.Key("feature"): t.String}
    )

    RowTrafaret = t.Dict(
        {
            t.Key("shap_explanation"): t.List(ShapContributionTrafaret) | t.Null,
            t.Key("timestamp"): t.String,
            t.Key("prediction"): t.Float,
        }
    )

    _converter = (
        t.Dict(
            {
                t.Key("count"): t.Int,
                t.Key("shap_base_value"): t.Float,
                t.Key("data"): t.List(RowTrafaret),
                t.Key("start_date"): t.String() | t.Null,
                t.Key("end_date"): t.String() | t.Null,
            }
        )
        .merge(RecordMetadataTrafaret)
        .ignore_extra("*")
    )

    def __init__(self, shap_base_value, data, start_date, end_date, count, **record_kwargs):
        self.shap_base_value = shap_base_value
        self.data = data
        self.count = count
        self.start_date = start_date
        self.end_date = end_date
        super(AnomalyAssessmentExplanations, self).__init__(**record_kwargs)

    @classmethod
    def get(cls, project_id, record_id, start_date=None, end_date=None, points_count=None):
        """Retrieve predictions along with shap explanations for the most anomalous records
        in the specified date range/for defined number of points.
        Two out of three parameters: start_date, end_date or points_count must be specified.

        Parameters
        ----------
        project_id: str
            The ID of the project.
        record_id: str
            The ID of the anomaly assessment record.
        start_date: str, optional
            The start of the date range to get explanations in.
            Example: ``2020-01-01T00:00:00.000000Z``
        end_date: str, optional
            The end of the date range to get explanations in.
            Example: ``2020-10-01T00:00:00.000000Z``
        points_count: int, optional
            The number of the rows to return.


        Returns
        -------
        AnomalyAssessmentExplanations

        """
        params = {}
        if start_date:
            params["startDate"] = start_date
        if end_date:
            params["endDate"] = end_date
        if points_count:
            params["pointsCount"] = points_count

        url = cls._path.format(project_id=project_id, record_id=record_id)
        r_data = cls._client.get(url, params=params).json()
        return cls.from_server_data(r_data)
